import requests, json, fnmatch, os, os.path, sys, subprocess, glob, ntpath, copy, re, operator, csv
from os import path
import gen3
from gen3.auth import Gen3Auth
from gen3.submission import Gen3Submission
from gen3.file import Gen3File

from pandas import json_normalize
import pandas as pd


class Gen3Error(Exception):
    pass


class Gen3Expansion:
    """Advanced scripts for interacting with the Gen3 submission, query and index APIs
    Args:
        endpoint (str): The URL of the data commons.
        auth_provider (Gen3Auth): A Gen3Auth class instance.

    Examples:
        This generates the Gen3Expansion class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> endpoint = "https://nci-crdc-demo.datacommons.io"
        ... auth = Gen3Auth(endpoint, refresh_file="credentials.json")
        ... exp = Gen3Expansion(endpoint, auth)

    """

    def __init__(self, endpoint, auth_provider):
        self._auth_provider = auth_provider
        self._endpoint = endpoint
        self.sub = Gen3Submission(endpoint, auth_provider)

    # Functions for downloading metadata in TSVs

    def get_project_ids(self, node=None, name=None):
        """Get a list of project_ids you have access to in a data commons.

        Args:
            node(str): The node you want projects to have at least one record in.
            name(str): The name of the programs to get projects in, or the submitter_id of a particular record.

        Example:
            get_project_ids()
            get_project_ids(node='demographic')
            get_project_ids(node='program',name='training')
            get_project_ids(node='case',name='case-01')
        """
        project_ids = []
        queries = []
        # Return all project_ids in the data commons if no node is provided or if node is program but no name provided
        if name is None and ((node is None) or (node is "program")):
            print("Getting all project_ids you have access to in the data commons.")
            if node == "program":
                print(
                    """Specify a list of program names (name = ['myprogram1','myprogram2']) to get only project_ids in particular programs."""
                )
            queries.append("""{project (first:0){project_id}}""")
        elif name is not None and node == "program":
            if isinstance(name, list):
                print(
                    """"Getting all project_ids in the programs '%s'"""
                    % (",".join(name))
                )
                for program_name in name:
                    queries.append(
                        """{project (first:0, with_path_to:{type:"program",name:"%s"}){project_id}}"""
                        % (program_name)
                    )
            elif isinstance(name, str):
                print("Getting all project_ids in the program '" + name + "'")
                queries.append(
                    """{project (first:0, with_path_to:{type:"program",name:"%s"}){project_id}}"""
                    % (name)
                )
        elif isinstance(node, str) and isinstance(name, str):
            print(
                """Getting all project_ids for projects with a path to record '%s' in node '%s'"""
                % (node, name)
            )
            queries.append(
                """{project (first:0, with_path_to:{type:"%s",submitter_id:"%s"}){project_id}}"""
                % (node, name)
            )
        elif isinstance(node, str) and name is None:
            print(
                """Getting all project_ids for projects with at least one record in the node '%s'"""
                % (node)
            )
            query = """{node (first:0,of_type:"%s"){project_id}}""" % (node)
            # Need separate logic because we are looking at res["data"]["node"]
            res = self.sub.query(query)
            new_project_ids = list(set([x["project_id"] for x in res["data"]["node"]]))
            project_ids = project_ids + new_project_ids

        if len(queries) > 0:
            for query in queries:
                res = self.sub.query(query)
                new_project_ids = list(
                    set([x["project_id"] for x in res["data"]["project"]])
                )
                project_ids = project_ids + new_project_ids
        my_ids = sorted(project_ids, key=str.lower)
        print(my_ids)
        return my_ids

    def get_node_tsvs(
        self,
        node,
        projects=None,
        overwrite=False,
        remove_empty=True,
        outdir="node_tsvs",
    ):
        """Gets a TSV of the structured data from particular node for each project specified.
           Also creates a master TSV of merged data from each project for the specified node.
           Returns a DataFrame containing the merged data for the specified node.

        Args:
            node (str): The name of the node to download structured data from.
            projects (list): The projects to download the node from. If "None", downloads data from each project user has access to.

        Example:
        >>> df = get_node_tsvs('demographic')

        """

        if not os.path.exists(outdir):
            os.makedirs(outdir)
        mydir = "{}/{}_tsvs".format(outdir, node)
        if not os.path.exists(mydir):
            os.makedirs(mydir)

        if projects is None:  # if no projects specified, get node for all projects
            projects = sorted(
                list(
                    set(
                        x["project_id"]
                        for x in self.sub.query("""{project (first:0){project_id}}""")[
                            "data"
                        ]["project"]
                    )
                )
            )
        elif isinstance(projects, str):
            projects = [projects]

        header = ""
        output = ""
        data_count = 0

        for project in projects:
            filename = str(mydir + "/" + project + "_" + node + ".tsv")
            if (os.path.isfile(filename)) and (overwrite is False):
                print("File previously downloaded.")
            else:
                prog, proj = project.split("-", 1)
                self.sub.export_node(prog, proj, node, "tsv", filename)

            with open(filename) as tsv_file:
                count = 0
                lines = tsv_file.readlines()
                # if file contains actual data and not just header row
                if len(lines) > 1:
                    for line in lines:
                        count += 1
                        # If header is empty and we are at the first line, set the header
                        if len(header) == 0 and count == 1:
                            header = line
                        elif count > 1:
                            output = output + line
                    count -= 1
                    data_count += count
                print(filename + " has " + str(count) + " records.")

                if remove_empty is True:
                    if count == 0:
                        print("Removing empty file: " + filename)
                        cmd = ["rm", filename]  # look in the download directory
                        try:
                            output = subprocess.check_output(
                                cmd, stderr=subprocess.STDOUT
                            ).decode("UTF-8")
                        except Exception as e:
                            output = e.output.decode("UTF-8")
                            print("ERROR deleting file: " + output)
        print("length of all data: " + str(data_count))
        nodefile = str("master_" + node + ".tsv")
        # Append header to the beginning of TSV
        output = header + output
        with open(nodefile, "w+") as master_tsv:
            master_tsv.write(output)
        print(
            "Master node TSV with "
            + str(data_count)
            + " total records written to "
            + nodefile
            + "."
        )
        return output

    def get_project_tsvs(
        self,
        projects=None,
        nodes=None,
        outdir="project_tsvs",
        overwrite=False,
        save_empty=False,
        remove_nodes=["program", "project", "root", "data_release"],
    ):
        """Function gets a TSV for every node in a specified project.
            Exports TSV files into a directory "project_tsvs/".
            Function returns a list of the contents of the directory.
        Args:
            projects (str/list): The project_id(s) of the project(s) to download. Can be a single project_id or a list of project_ids.
            nodes(str/list): The nodes to download from each project. If None, will try to download all nodes in the data model.
            overwrite (boolean): If False, the TSV file is not downloaded if there is an existing file with the same name.
            save_empty(boolean): If True, TSVs with no records, i.e., downloads an empty TSV template, will be downloaded.
            remove_nodes(list): A list of nodes in the data model that should not be downloaded per project.
        Example:
        >>> get_project_tsvs(projects = ['internal-test'])

        """
        if nodes is None:
            # get all the 'node_id's in the data model
            nodes = sorted(
                list(
                    set(
                        x["id"]
                        for x in self.sub.query("""{_node_type (first:-1) {id}}""")[
                            "data"
                        ]["_node_type"]
                    )
                )
            )
        elif isinstance(nodes, str):
            nodes = [nodes]

        for node in remove_nodes:
            if node in nodes:
                nodes.remove(node)

        if projects is None:  # if no projects specified, get node for all projects
            projects = sorted(
                list(
                    set(
                        x["project_id"]
                        for x in self.sub.query("""{project (first:0){project_id}}""")[
                            "data"
                        ]["project"]
                    )
                )
            )
        elif isinstance(projects, str):
            projects = [projects]

        for project_id in projects:
            mydir = "{}/{}_tsvs".format(
                outdir, project_id
            )  # create the directory to store TSVs

            if not os.path.exists(mydir):
                os.makedirs(mydir)

            for node in nodes:
                filename = str(mydir + "/" + project_id + "_" + node + ".tsv")
                if (os.path.isfile(filename)) and (overwrite is False):
                    print("\tPreviously downloaded: '{}'".format(filename))
                else:
                    query_txt = """{_%s_count (project_id:"%s")}""" % (node, project_id)
                    res = self.sub.query(
                        query_txt
                    )  #  {'data': {'_acknowledgement_count': 0}}
                    count = res["data"][str("_" + node + "_count")]  # count=int(0)
                    if count > 0 or save_empty is True:
                        print(
                            "\nDownloading {} records in node '{}' of project '{}'.".format(
                                count, node, project_id
                            )
                        )
                        prog, proj = project_id.split("-", 1)
                        self.sub.export_node(prog, proj, node, "tsv", filename)
                    else:
                        print(
                            "\t{} records in node '{}' of project '{}'.".format(
                                count, node, project_id
                            )
                        )

        cmd = ["ls", mydir]  # look in the download directory
        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode(
                "UTF-8"
            )
        except Exception as e:
            output = "ERROR:" + e.output.decode("UTF-8")

        return output



"""
Module for metadata actions for verifying a manifest of
metadata objects (against mds's API). Supports
multiple processes and coroutines using Python's asyncio library.

Manifest must include a guid column in addition to an arbitrary number of columns
representing expected metadata fields. Fields that are key:value mappings should be
valid JSON strings.

There is a default mapping for getting the GUID and metadata fields but you can override it.
Fields that expect lists (like acl, authz, and urls) by default assume these values are
separated with spaces. If you need alternate behavior, you can simply override the
`manifest_row_parsers`.

The output from this verification is a file containing any errors in the following
format:

{guid}|{mismatched_field}|expected {value_from_manifest}|actual {value_from_mds}

Attributes:
    CURRENT_DIR (str): directory this file is in
    MAX_CONCURRENT_REQUESTS (int): maximum number of desired concurrent requests across
        processes/threads
"""
import asyncio
import aiohttp
import csv
from collections import OrderedDict
from collections.abc import Mapping
import json
from cdislogging import get_logger

import os
import time

from gen3.metadata import Gen3Metadata
from gen3.utils import get_or_create_event_loop_for_thread

MAX_CONCURRENT_REQUESTS = 24
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

logging = get_logger("__name__")


def _get_guid_from_row(row):
    """
    Given a row from the manifest, return the field representing expected mds guid.

    Args:
        row (dict): column_name:row_value

    Returns:
        str: guid
    """
    guid = row.get("guid")
    if not guid:
        guid = row.get("GUID")
    return guid


def _get_metadata_from_row(row):
    """
    Given a row from the manifest, return the field representing metadata.

    Args:
        row (dict): column_name:row_value

    Returns:
    """
    metadata = dict(row)

    # make sure guid is not part of the metadata
    if "guid" in metadata:
        del metadata["guid"]

    if "GUID" in metadata:
        del metadata["GUID"]

    return metadata


manifest_row_parsers = {"guid": _get_guid_from_row, "metadata": _get_metadata_from_row}


async def async_verify_metadata_manifest(
    commons_url,
    manifest_file,
    metadata_source,
    max_concurrent_requests=MAX_CONCURRENT_REQUESTS,
    manifest_row_parsers=manifest_row_parsers,
    manifest_file_delimiter=None,
    output_filename=f"verify-metadata-errors-{time.time()}.log",
):
    """
    Verify all file object records

    Args:
        commons_url (str): root domain for commons where mds lives
        manifest_file (str): the file to verify against
        metadata_source (str): the source of the metadata you are verifying, in practice
            this means the first nested section in the metadata service
        max_concurrent_requests (int): the maximum number of concurrent requests allowed
        manifest_row_parsers (Dict{mds_field:func_to_parse_row}): Row parsers
        manifest_file_delimiter (str): delimeter in manifest_file
        output_filename (str): filename for output logs
    """
    start_time = time.perf_counter()
    logging.info(f"start time: {start_time}")

    # if delimiter not specified, try to get based on file ext
    if not manifest_file_delimiter:
        file_ext = os.path.splitext(manifest_file)
        if file_ext[-1].lower() == ".tsv":
            manifest_file_delimiter = "\t"
        else:
            # default, assume CSV
            manifest_file_delimiter = ","

    await _verify_all_metadata_records_in_file(
        commons_url,
        manifest_file,
        manifest_file_delimiter,
        max_concurrent_requests,
        output_filename,
        metadata_source,
    )

    end_time = time.perf_counter()
    logging.info(f"end time: {end_time}")
    logging.info(f"run time: {end_time-start_time}")


async def _verify_all_metadata_records_in_file(
    commons_url,
    manifest_file,
    manifest_file_delimiter,
    max_concurrent_requests,
    output_filename,
    metadata_source,
):
    """
    Getting mds records and write output to a file. This function
    creates semaphores to limit the number of concurrent http connections that
    get opened to send requests to mds.

    It then uses asyncio to start a number of coroutines. Steps:
        1) requests to mds to get records (writes resulting records to a queue)
        2) reading those records from the queue and writing to a file

    Args:
        commons_url (str): root domain for commons where mds lives
        manifest_file (str): the file to verify against
        manifest_file_delimiter (str): delimeter in manifest_file
        output_filename (str, optional): filename for output
        max_concurrent_requests (int): the maximum number of concurrent requests allowed
        metadata_source (str): the source of the metadata you are verifying, in practice
            this means the first nested section in the metadata service
    """
    max_requests = int(max_concurrent_requests)
    logging.debug(f"max concurrent requests: {max_requests}")
    lock = asyncio.Semaphore(max_requests)
    queue = asyncio.Queue()
    output_queue = asyncio.Queue()

    with open(manifest_file, encoding="utf-8-sig") as manifest:
        reader = csv.DictReader(manifest, delimiter=manifest_file_delimiter)
        for row in reader:
            new_row = {}
            for key, value in row.items():
                new_row[key.strip()] = value.strip()
            await queue.put(new_row)

    await asyncio.gather(
        *(
            _parse_from_queue(queue, lock, commons_url, output_queue, metadata_source)
            # why "+ (max_concurrent_requests / 4)"?
            # This is because the max requests at any given time could be
            # waiting for metadata responses all at once and there's processing done
            # before that semaphore, so this just adds a few extra processes to get
            # through the queue up to that point of metadata requests so it's ready
            # right away when a lock is released. Not entirely necessary but speeds
            # things up a tiny bit to always ensure something is waiting for that lock
            for x in range(
                0, int(max_concurrent_requests + (max_concurrent_requests / 4))
            )
        )
    )

    output_filename = os.path.abspath(output_filename)
    logging.info(
        f"done processing, writing output queue to single file {output_filename}"
    )

    # remove existing output if it exists
    if os.path.isfile(output_filename):
        os.unlink(output_filename)

    with open(output_filename, "w") as outfile:
        while not output_queue.empty():
            line = await output_queue.get()
            outfile.write(line)

    logging.info(f"done writing output to file {output_filename}")


async def _parse_from_queue(queue, lock, commons_url, output_queue, metadata_source):
    """
    Keep getting items from the queue and verifying that mds contains the expected
    fields from that row. If there are any issues, log errors into a file. Return
    when nothing is left in the queue.

    Args:
        queue (asyncio.Queue): queue to read mds records from
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
        commons_url (str): root domain for commons where mds lives
        output_queue (asyncio.Queue): queue for output
        metadata_source (str): the source of the metadata you are verifying, in practice
            this means the first nested section in the metadata service
    """
    loop = get_or_create_event_loop_for_thread()

    while not queue.empty():
        row = await queue.get()

        guid = manifest_row_parsers["guid"](row)
        metadata = manifest_row_parsers["metadata"](row)

        actual_record = await _get_record_from_mds(guid, commons_url, lock)
        if not actual_record:
            output = f"{guid}|no_record|expected {row}|actual None\n"
            await output_queue.put(output)
            logging.error(output)
        else:
            logging.info(f"verifying {guid}...")

            for expected_key, expected_value in metadata.items():
                actual_value_raw = actual_record.get(metadata_source, {}).get(
                    expected_key
                )
                try:
                    actual_value = json.loads(str(actual_value_raw))
                except json.decoder.JSONDecodeError as exc:
                    actual_value = actual_value_raw

                try:
                    expected_value = json.loads(str(expected_value))
                except json.decoder.JSONDecodeError as exc:
                    pass

                # compare as dicts if necessary, otherwise just compare values
                if (
                    isinstance(expected_value, Mapping)
                    and not _are_matching_dicts(expected_value, actual_value)
                ) or (actual_value != expected_value):
                    output = f"{guid}|{metadata_source}.{expected_key}|expected {expected_value}|actual {actual_value}\n"
                    await output_queue.put(output)
                    logging.error(output)


def _are_matching_dicts(dict_a, dict_b):
    # if one of these isn't a dict, just try a normal comparison
    if not getattr(dict_a, "keys", None) or not getattr(dict_b, "keys", None):
        return dict_a == dict_b

    if len(dict_a.keys()) != len(dict_b.keys()):
        return False

    for key_a, value_a in dict_a.items():
        value_b = dict_b.get(key_a)
        if value_a != value_b:
            return False

    return True


async def _get_record_from_mds(guid, commons_url, lock):
    """
    Gets a semaphore then requests a record for the given guid

    Args:
        guid (str): mds record globally unique id
        commons_url (str): root domain for commons where mds lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    mds = Gen3Metadata(commons_url)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        try:
            return await mds.async_get(guid, _ssl=ssl)
        except aiohttp.client_exceptions.ClientResponseError:
            return None



import click


from gen3.tools.wrap import Gen3Wrap


@click.command(
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
    help="A wrapper command that forwards COMMAND_ARGS as-is after setting the environment variable GEN3_TOKEN",
)
@click.argument("command_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def run(ctx, command_args):
    auth = ctx.obj["auth_factory"].get()
    gen3Wrap_object = Gen3Wrap(auth, command_args)
    gen3Wrap_object.run_command()



import click

from gen3users import main as users_cli


try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points


@click.group()
def main():
    """Gen3 Command Line Interface"""
    pass


@click.group()
def users():
    """Commands for working with gen3users"""
    pass


for command in users_cli.main.commands:
    users.add_command(users_cli.main.get_command(ctx=None, cmd_name=command))

# load plug-ins from entry_points
for ep in entry_points().get("gen3.plugins", []):
    ep.load()



"""
Module for indexing actions for verifying a manifest of
indexed file objects (against indexd's API). Supports
multiple processes and coroutines using Python's asyncio library.

The default manifest format created is a Comma-Separated Value file (csv)
with rows for every record. A header row is created with field names:
guid,authz,acl,file_size,md5,urls,file_name

Fields that are lists (like acl, authz, and urls) separate the values with spaces.

There is a default mapping for those column names above but you can override it.
Fields that expect lists (like acl, authz, and urls) by default assume these values are
separated with spaces. If you need alternate behavior, you can simply override the
`manifest_row_parsers` for specific fields and replace the default parsing function
with a custom one. For example:

```
from gen3.tools import indexing
from gen3.tools.indexing.verify_manifest import manifest_row_parsers

def _get_authz_from_row(row):
    return [row.get("authz").strip().strip("[").strip("]").strip("'")]

# override default parsers
manifest_row_parsers["authz"] = _get_authz_from_row

indexing.verify_object_manifest(COMMONS)
```

The output from this verification is a file containing any errors in the following
format:

{guid}|{error_name}|expected {value_from_manifest}|actual {value_from_indexd}
ex: 93d9af72-b0f1-450c-a5c6-7d3d8d2083b4|authz|expected ['']|actual ['/programs/DEV/projects/test']

Attributes:
    CURRENT_DIR (str): directory this file is in
    MAX_CONCURRENT_REQUESTS (int): maximum number of desired concurrent requests across
        processes/threads
"""
import aiohttp
import asyncio
import csv
from cdislogging import get_logger

import os
import time

from gen3.index import Gen3Index
from gen3.utils import get_or_create_event_loop_for_thread

MAX_CONCURRENT_REQUESTS = 24
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

logging = get_logger("__name__")


def _get_guid_from_row(row):
    """
    Given a row from the manifest, return the field representing expected indexd guid.

    Args:
        row (dict): column_name:row_value

    Returns:
        str: guid
    """
    guid = row.get("guid")
    if not guid:
        guid = row.get("GUID")
    return guid


def _get_md5_from_row(row):
    """
    Given a row from the manifest, return the field representing file's md5 sum.

    Args:
        row (dict): column_name:row_value

    Returns:
        str: md5 sum for file
    """
    if "md5" in row:
        return row["md5"]
    elif "md5sum" in row:
        return row["md5sum"]
    else:
        return None


def _get_file_size_from_row(row):
    """
    Given a row from the manifest, return the field representing file's size in bytes.

    Args:
        row (dict): column_name:row_value

    Returns:
        int: integer representing file size in bytes
    """
    try:
        if "file_size" in row:
            return int(row["file_size"])
        elif "size" in row:
            return int(row["size"])
        else:
            return None
    except Exception:
        logging.warning(f"could not convert this to an int: {row.get('file_size')}")
        return row.get("file_size")


def _get_acl_from_row(row):
    """
    Given a row from the manifest, return the field representing file's expected acls.

    Args:
        row (dict): column_name:row_value

    Returns:
        List[str]: acls for the indexd record
    """
    return [item for item in row.get("acl", "").strip().split(" ") if item]


def _get_authz_from_row(row):
    """
    Given a row from the manifest, return the field representing file's expected authz
    resources.

    Args:
        row (dict): column_name:row_value

    Returns:
        List[str]: authz resources for the indexd record
    """
    return [item for item in row.get("authz", "").strip().split(" ") if item]


def _get_urls_from_row(row):
    """
    Given a row from the manifest, return the field representing file's expected urls.

    Args:
        row (dict): column_name:row_value

    Returns:
        List[str]: urls for indexd record file location(s)
    """
    if "urls" in row:
        return [item for item in row.get("urls", "").strip().split(" ") if item]
    elif "url" in row:
        return [item for item in row.get("urls", "").strip().split(" ") if item]
    else:
        return []


def _get_file_name_from_row(row):
    """
    Given a row from the manifest, return the field representing file's expected file_name.

    Args:
        row (dict): column_name:row_value

    Returns:
        List[str]: file_name for indexd record file location(s)
    """
    if "file_name" in row:
        return row["file_name"]
    elif "filename" in row:
        return row["filename"]
    elif "name" in row:
        return row["name"]
    else:
        return None


manifest_row_parsers = {
    "guid": _get_guid_from_row,
    "md5": _get_md5_from_row,
    "file_size": _get_file_size_from_row,
    "acl": _get_acl_from_row,
    "authz": _get_authz_from_row,
    "urls": _get_urls_from_row,
    "file_name": _get_file_name_from_row,
}


async def async_verify_object_manifest(
    commons_url,
    manifest_file,
    max_concurrent_requests=MAX_CONCURRENT_REQUESTS,
    manifest_row_parsers=manifest_row_parsers,
    manifest_file_delimiter=None,
    output_filename=f"verify-manifest-errors-{time.time()}.log",
):
    """
    Verify all file object records into a manifest csv

    Args:
        commons_url (str): root domain for commons where indexd lives
        manifest_file (str): the file to verify against
        max_concurrent_requests (int): the maximum number of concurrent requests allowed
        manifest_row_parsers (Dict{indexd_field:func_to_parse_row}): Row parsers
        manifest_file_delimiter (str): delimeter in manifest_file
        output_filename (str): filename for output logs
    """
    start_time = time.perf_counter()
    logging.info(f"start time: {start_time}")

    # if delimiter not specified, try to get based on file ext
    if not manifest_file_delimiter:
        file_ext = os.path.splitext(manifest_file)
        if file_ext[-1].lower() == ".tsv":
            manifest_file_delimiter = "\t"
        else:
            # default, assume CSV
            manifest_file_delimiter = ","

    logging.debug(f"detected {manifest_file_delimiter} as delimiter between columns")

    await _verify_all_index_records_in_file(
        commons_url,
        manifest_file,
        manifest_file_delimiter,
        max_concurrent_requests,
        output_filename,
    )

    end_time = time.perf_counter()
    logging.info(f"end time: {end_time}")
    logging.info(f"run time: {end_time-start_time}")


async def _verify_all_index_records_in_file(
    commons_url,
    manifest_file,
    manifest_file_delimiter,
    max_concurrent_requests,
    output_filename,
):
    """
    Getting indexd records and writing to a file. This function
    creates semaphores to limit the number of concurrent http connections that
    get opened to send requests to indexd.

    It then uses asyncio to start a number of coroutines. Steps:
        1) requests to indexd to get records (writes resulting records to a queue)
        2) puts a final "DONE" in the queue to stop coroutine that will read from queue
        3) reading those records from the queue and writing to a file

    Args:
        commons_url (str): root domain for commons where indexd lives
        manifest_file (str): the file to verify against
        manifest_file_delimiter (str): delimeter in manifest_file
        output_filename (str, optional): filename for output
        max_concurrent_requests (int): the maximum number of concurrent requests allowed
    """
    max_requests = int(max_concurrent_requests)
    logging.debug(f"max concurrent requests: {max_requests}")
    lock = asyncio.Semaphore(max_requests)
    queue = asyncio.Queue()
    output_queue = asyncio.Queue()

    with open(manifest_file, encoding="utf-8-sig") as manifest:
        reader = csv.DictReader(manifest, delimiter=manifest_file_delimiter)

        for row in reader:
            new_row = {}
            for key, value in row.items():
                if value:
                    value = value.strip()

                new_row[key.strip()] = value
            await queue.put(new_row)

    for _ in range(0, int(max_concurrent_requests + (max_concurrent_requests / 4))):
        await queue.put("DONE")

    await asyncio.gather(
        *(
            _parse_from_queue(queue, lock, commons_url, output_queue)
            # why "+ (max_concurrent_requests / 4)"?
            # This is because the max requests at any given time could be
            # waiting for metadata responses all at once and there's processing done
            # before that semaphore, so this just adds a few extra processes to get
            # through the queue up to that point of metadata requests so it's ready
            # right away when a lock is released. Not entirely necessary but speeds
            # things up a tiny bit to always ensure something is waiting for that lock
            for x in range(
                0, int(max_concurrent_requests + (max_concurrent_requests / 4))
            )
        )
    )

    output_filename = os.path.abspath(output_filename)
    logging.info(
        f"done processing, writing output queue to single file {output_filename}"
    )

    # remove existing output if it exists
    if os.path.isfile(output_filename):
        os.unlink(output_filename)

    with open(output_filename, "w") as outfile:
        while not output_queue.empty():
            line = await output_queue.get()
            outfile.write(line)

    logging.info(f"done writing output to file {output_filename}")


async def _parse_from_queue(queue, lock, commons_url, output_queue):
    """
    Keep getting items from the queue and verifying that indexd contains the expected
    fields from that row. If there are any issues, log errors into a file. Return
    when nothing is left in the queue.

    Args:
        queue (asyncio.Queue): queue to read indexd records from
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
        commons_url (str): root domain for commons where indexd lives
        output_queue (asyncio.Queue): queue for output
    """
    loop = get_or_create_event_loop_for_thread()

    row = await queue.get()

    while row != "DONE":
        guid = manifest_row_parsers["guid"](row)
        authz = manifest_row_parsers["authz"](row)
        acl = manifest_row_parsers["acl"](row)
        file_size = manifest_row_parsers["file_size"](row)
        md5 = manifest_row_parsers["md5"](row)
        urls = manifest_row_parsers["urls"](row)
        file_name = manifest_row_parsers["file_name"](row)

        actual_record = await _get_record_from_indexd(guid, commons_url, lock)
        if not actual_record:
            output = f"{guid}|no_record|expected {row}|actual None\n"
            await output_queue.put(output)
            logging.error(output)
        else:
            logging.info(f"verifying {guid}...")

            if sorted(authz) != sorted(actual_record["authz"]):
                output = (
                    f"{guid}|authz|expected {authz}|actual {actual_record['authz']}\n"
                )
                await output_queue.put(output)
                logging.error(output)

            if sorted(acl) != sorted(actual_record["acl"]):
                output = f"{guid}|acl|expected {acl}|actual {actual_record['acl']}\n"
                await output_queue.put(output)
                logging.error(output)

            if file_size != actual_record["size"]:
                if (
                    not file_size
                    and file_size != 0
                    and not actual_record["size"]
                    and actual_record["size"] != 0
                ):
                    # actual and expected are both either empty string or None
                    # so even though they're not equal, they represent null value so
                    # we don't need to consider this an error in validation
                    pass
                else:
                    output = f"{guid}|file_size|expected {file_size}|actual {actual_record['size']}\n"
                    await output_queue.put(output)
                    logging.error(output)

            if md5 != actual_record["hashes"].get("md5"):
                if (
                    not md5
                    and md5 != 0
                    and not actual_record["hashes"].get("md5")
                    and actual_record["hashes"].get("md5") != 0
                ):
                    # actual and expected are both either empty string or None
                    # so even though they're not equal, they represent null value so
                    # we don't need to consider this an error in validation
                    pass
                else:
                    output = f"{guid}|md5|expected {md5}|actual {actual_record['hashes'].get('md5')}\n"
                    await output_queue.put(output)
                    logging.error(output)
            urls = [url.replace("%20", " ") for url in urls]
            if sorted(urls) != sorted(actual_record["urls"]):
                output = f"{guid}|urls|expected {urls}|actual {actual_record['urls']}\n"
                await output_queue.put(output)
                logging.error(output)

            if not actual_record["file_name"] and file_name:
                # if the actual record name is "" or None but something was specified
                # in the manifest, we have a problem
                output = f"{guid}|file_name|expected {file_name}|actual {actual_record['file_name']}\n"
                await output_queue.put(output)
                logging.error(output)

        row = await queue.get()


async def _get_record_from_indexd(guid, commons_url, lock):
    """
    Gets a semaphore then requests a record for the given guid

    Args:
        guid (str): indexd record globally unique id
        commons_url (str): root domain for commons where indexd lives
        lock (asyncio.Semaphore): semaphones used to limit ammount of concurrent http
            connections
    """
    index = Gen3Index(commons_url)
    async with lock:
        # default ssl handling unless it's explicitly http://
        ssl = None
        if "https" not in commons_url:
            ssl = False

        record = None

        try:
            return await index.async_get_record(guid, _ssl=ssl)

        except aiohttp.client_exceptions.ClientResponseError as exc:
            logging.warning(f"couldn't get record. error: {exc}")

        return record



import requests
from gen3.utils import raise_for_status_and_print_error


class Gen3ObjectError(Exception):
    pass


class Gen3Object:
    """For interacting with Gen3 object level features.

    A class for interacting with the Gen3 object services.
    Currently allows creating and deleting of an object from the Gen3 System.

    Args:
        auth_provider (Gen3Auth): A Gen3Auth class instance.

    Examples:
        This generates the Gen3Object class pointed at the sandbox commons while
        using the credentials.json downloaded from the commons profile page.

        >>> auth = Gen3Auth(refresh_file="credentials.json")
        ... object = Gen3Object(auth)

    """

    def __init__(self, auth_provider=None):
        self._auth_provider = auth_provider
        self.service_endpoint = "/mds"

    def create_object(self, file_name, authz, metadata=None, aliases=None):
        url = (
            self._auth_provider.endpoint.rstrip("/")
            + self.service_endpoint
            + "/objects"
        )
        body = {
            "file_name": file_name,
            "authz": authz,
            "metadata": metadata,
            "aliases": aliases,
        }
        response = requests.post(url, json=body, auth=self._auth_provider)
        raise_for_status_and_print_error(response)
        data = response.json()
        return data["guid"], data["upload_url"]

    def delete_object(self, guid, delete_file_locations=False):
        """
        Delete the object from indexd, metadata service and optionally all storage locations

        Args:
            `guid`  -- GUID of the object to delete
            `delete_file_locations` -- if True, removes the object from existing bucket location(s) through fence
        Returns:
            Nothing
        """
        delete_param = "?delete_file_locations" if delete_file_locations else ""
        url = (
            self._auth_provider.endpoint.rstrip("/")
            + self.service_endpoint
            + "/objects/"
            + guid
            + delete_param
        )
        response = requests.delete(url, auth=self._auth_provider)
        raise_for_status_and_print_error(response)



import requests

from cdislogging import get_logger
import xmltodict

from gen3.external import ExternalMetadataSourceInterface
from gen3.external.nih.utils import get_dbgap_accession_as_parts
from gen3.utils import append_query_params

logging = get_logger("__name__")


class dbgapStudyRegistration(ExternalMetadataSourceInterface):
    """
    Class for interacting with National Center for
    Biotechnology Information (NCBI)'s database of Genotypes and Phenotypes (dbGaP)
    Study Registration API. The Study Registration API contains
    public information about studies that are registered in dbGaP with things like
    who the PI is, what is the funding institution, what child studies are there,
    etc.

    Example Usage:

        ```
        from gen3.external.nih.dbgap_study_registration import dbgapStudyRegistration


        def main():
            studies = [
                "phs000007.v32",
                "phs000179.v1.p1.c1",
            ]
            dbgap_study_registration = dbgapStudyRegistration()
            metadata = dbgap_study_registration.get_metadata_for_ids(ids=studies)
            print(metadata)


        if __name__ == "__main__":
            main()
        ```
    """

    def __init__(
        self,
        api="https://dbgap.ncbi.nlm.nih.gov/ss/dbgapssws.cgi",
        auth_provider=None,
    ):
        super(dbgapStudyRegistration, self).__init__(api, auth_provider)

    def get_metadata_for_ids(self, ids):
        """
        Return dbGaP Study Registration metadata for each of the provided IDs.

        Args:
            ids (List[str]): list of IDs to query for

        Returns:
            Dict[dict]: metadata for each of the provided IDs (which
                        are the keys in the returned dict)
        """
        logging.info("Getting dbGaP Study Registration metadata...")
        logging.debug("Provided ids for dbGaP Study Registration metadata: {ids}")

        metadata_for_ids = {}

        for phsid in ids:
            standardized_phsid_parts = get_dbgap_accession_as_parts(phsid)

            if not standardized_phsid_parts["version_number"]:
                logging.warning(
                    f"ID provided '{phsid}' does not specify a version. This will "
                    f"still return data without specifying the version. However, "
                    f"ensure the version returned is appropriate. The default "
                    f"version when not specified is usually the latest version in dbGaP."
                )

            request_url = append_query_params(
                self.api,
                request="Study",
                phs=standardized_phsid_parts["phsid_number"],
            )
            if standardized_phsid_parts["version_number"]:
                request_url = append_query_params(
                    request_url,
                    v=standardized_phsid_parts["version_number"],
                )

            logging.debug(f"GET: {request_url}")
            raw_xml_response = requests.get(request_url)
            logging.debug(
                f"Got XML response: {raw_xml_response.text}, converting to JSON..."
            )

            # this library has an opinionated way to convert XML to JSON that
            # preserves attributes on tags by appending them with '@'.
            # The parsing follows this standard:
            #   https://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html
            try:
                response = xmltodict.parse(raw_xml_response.text)
            except Exception as exc:
                logging.error(
                    f"{phsid} has an invalid dbGaP Study Registration metadata "
                    "response. Cannot continue without it. Skipping..."
                )
                continue

            logging.debug(f"Got full JSON: {response}")

            study = response.get("dbgapss", {}).get("Study", [])
            if type(study) == list:
                if len(study) > 1:
                    logging.warning(
                        f"Multiple studies found from {request_url}, using first one ONLY..."
                    )
                    study = study[0]
                elif len(study) == 0:
                    logging.warning(f"No study found from {request_url}. Skipping...")
                    continue

            logging.debug(f"Got study JSON: {study}")

            metadata_for_ids[phsid] = study

        return metadata_for_ids

    def get_child_studies_for_ids(self, ids):
        """
        Returns the dbGaP child studies, if any, for each of the provided dbGaP study IDs.

          Args:
            ids (List[str]): list of IDs to query for

        Returns:
            Dict[dict]: list of children IDs for each of the provided IDs (which
                        are the keys in the returned dict)
        """
        parents = self.get_metadata_for_ids(ids)
        result = {}
        for parent, metadata in parents.items():
            children = metadata["StudyInfo"].get("childAccession", [])
            # dbGaP returns a string if there's only one child, so force it into a list here so the dict is consistent
            if isinstance(metadata["StudyInfo"].get("childAccession", []), list):
                result[parent] = children
            else:
                result[parent] = [children]
        return result

    def get_parent_studies_for_ids(self, ids):
        """
        Returns the dbGaP parent study, if any, for each of the provided dbGaP study IDs.

          Args:
            ids (List[str]): list of IDs to query for

        Returns:
            Dict[str]: dictionary mapping the child study IDs to their parent study IDs. If a given study ID does not
            have a parent study, then the studyId will map to None

        """
        children = self.get_metadata_for_ids(ids)
        result = {}
        for child, metadata in children.items():
            parent = metadata["StudyInfo"].get("@parentAccession", None)

            # the input ids may not include a version or partipant set, so we also
            # want to compare the bare minimum "phs" with the input
            #
            # this will now correctly handle cases where the child ID is
            # "phs001173"
            # and the parent returned is "phs001173.v1.p1" (which really
            # are the same)
            standardized_phsid_parts = get_dbgap_accession_as_parts(parent)
            parent_simplified = standardized_phsid_parts["phsid"]

            if child == parent or child == parent_simplified:
                result[child] = None
            else:
                result[child] = parent
        return result



"""
Module for downloading and listing data from external repositories.

    Examples:

        See docs/howto/externalFileDownloading.md for more details

"""
import json
from pathlib import Path
from typing import Any, Dict, List

from cdislogging import get_logger

from gen3.auth import Gen3Auth

REQUIRED_EXTERNAL_FILE_METADATA_KEYS = ["file_retriever"]

logger = get_logger("__name__", log_level="debug")


def download_files_from_metadata(
    hostname: str,
    auth: Gen3Auth,
    external_file_metadata: Dict,
    retrievers: Dict,
    download_path: str = ".",
) -> Dict[str, Any]:
    """
    Download data from an external repository using
    external-file metadata and a retriever function.

    The external_file_metadata items should have the following
    required key: 'file_retriever'.
    The items can have an additional optional key of
    'study_id' or 'file_id.

    An example is shown below:
    [
        {
            "external_oidc_idp": "externaldata-keycloak",
            "file_retriever": "QDR",
            "study_id": "QDR_study_01"
        },
        {
            "external_oidc_idp": "externaldata-keycloak",
            "file_retriever": "QDR",
            "file_id": "QDR_file_02"
        },
        {
            "file_retriever": "Dataverse",
            "file_id": "Dataverse_file_01"
        },
    ]

    JSON does not allow for storing references to the retriever functions. We store
    a string for 'file_retriever' value as shown above. The retriever name string
    will get mapped to a function.

    For a retriever name of 'QDR', the mapping to a function could be as follows:
    { "QDR": get_syracuse_qdr_files }

    The retriever function should be imported into the code that is calling
    the 'download_files_from_metadata' function.

    The retriever function (not defined here) will perform the data download.

    Args:
        hostname (str): hostname of Gen3 commons to use for WTS
        auth: Gen3Auth instance for WTS
        external_file_metadata (dict): dict of external file objects
        retrievers (dict): map of function name string to retreiver function reference.
    Returns:
        dictionary of DownloadStatus for external file objects, None if errors
    """

    if retrievers == {}:
        logger.critical(f"ERROR: 'retrievers' specification is empty {retrievers}")
        return None

    if external_file_metadata == None:
        logger.critical(f"ERROR: external_file_metadata is missing")
        return None

    download_status = pull_files(
        wts_server_name=hostname,
        auth=auth,
        file_metadata_list=external_file_metadata,
        retrievers=retrievers,
        download_path=download_path,
    )

    if download_status == None:
        logger.critical("Download status is None")
    else:
        for key in download_status:
            logger.info(download_status[key])

    return download_status


def pull_files(
    wts_server_name: str,
    auth,
    file_metadata_list,
    retrievers,
    download_path,
) -> Dict[str, Any]:
    """
    Pull data files from external repo with external function.

    Args:
        wts_server_name (str): hostname of WTS server
        auth: Gen3Auth object for access to WTS
        file_metadata_list (list): list of external file metadata objects
        retrievers (dict):
        download_path: path to write data files
    Returns:
        dictionary of DownloadStatus for external file objects, None if errors
    """
    completed = {}

    metadata_grouped_by_retriever = check_data_and_group_by_retriever(
        file_metadata_list
    )

    if metadata_grouped_by_retriever == None:
        logger.critical(f"ERROR: Cannot group external metadata by retrievers")
        return None
    logger.debug(f"Grouped metadata = {metadata_grouped_by_retriever}")

    for key, val in metadata_grouped_by_retriever.items():
        retriever_function = retrievers.get(key)
        if retriever_function == None:
            logger.critical(f"Could not find retriever function for {key}")
            continue
        file_metadata_list = val
        download_status = None

        try:
            logger.info(
                f"Ready to retrieve data with retriever {retriever_function.__name__}"
            )
            download_status = retriever_function(
                wts_server_name, auth, file_metadata_list, download_path
            )
        except NameError:
            logger.critical("ERROR: Retriever function not in scope.")
        except Exception as e:
            logger.critical("Error from retriever function")
            logger.critical(e)

        if download_status:
            logger.info(f"Adding download_status = {download_status}")
            completed.update(download_status)

    if completed == {}:
        return None
    else:
        return completed


def extract_external_file_metadata(mds_object: dict) -> List:
    """
    Extract the list of dictionaries of external_file_metadata from an mds object.
    Return None if the external_file_metadata is not present.

    Args:
        mds_object (dict): the metadata from the input file
    Returns:
        List of external_file_metadata dictionaries
    """
    try:
        external_file_metadata = mds_object.get("gen3_discovery").get(
            "external_file_metadata", None
        )
    except AttributeError as e:
        logger.critical("ERROR: Missing key 'gen3_discovery' in metadata")
        logger.critical(e)
        return None

    return external_file_metadata


def is_valid_external_file_metadata(external_file_metadata: dict) -> bool:
    """
    Check that the external_file dict has the required key(s) as defined in 'REQUIRED_EXTERNAL_FILE_METADATA_KEYS'

    Args:
        external_file_metadata: dict
    Returns:
        boolean
    """
    if not isinstance(external_file_metadata, dict):
        logger.info(f"File_metadata object is not a dict: {external_file_metadata}")
        return False

    isValid = True
    logger.debug(f"Checking metadata: {external_file_metadata}")
    for key in REQUIRED_EXTERNAL_FILE_METADATA_KEYS:
        if key not in external_file_metadata:
            logger.info(f"Missing key in external file metadata: {key}")
            isValid = False

    return isValid


def load_all_metadata(path: Path) -> Dict:
    """
    Loads a metadata object from a json file.

    Args:
        path (Path): Path to metadata json file
    Returns:
        dict of metadata, None if failure
    """
    try:
        with open(path, "rt") as f:
            metadata = json.load(f)
            return metadata
    except FileNotFoundError as ex:
        logger.critical(f"Error file not found: {ex.filename}")
    except json.JSONDecodeError as ex:
        logger.critical(f"format of manifest file is valid JSON: {ex.msg}")
    return None


def check_data_and_group_by_retriever(file_metadata_list: list) -> Dict:
    """
    Group the file metadata by retriever. Exclude any invalid file_metadata items.

    Returns:
        Dict of retrievers and file_metadata lists, for example
        {
            "retriever1": [list1],
            "retriever2": [list2],
            ...
        }
    """
    grouped_by_retriever = {}
    for item in file_metadata_list:
        logger.debug(f"Checking item = {item}")
        if not is_valid_external_file_metadata(item):
            continue
        retriever = item.get("file_retriever")
        logger.debug(f"Found retriever = {retriever}")
        if retriever in grouped_by_retriever:
            grouped_by_retriever[retriever].append(item)
        else:
            grouped_by_retriever[retriever] = [item]

    if grouped_by_retriever == {}:
        return None
    return grouped_by_retriever



import csv
import json
import tempfile
import asyncio
import os
import requests.exceptions
from cdislogging import get_logger
from urllib.parse import urlparse

from gen3.metadata import Gen3Metadata

from gen3.tools.metadata.discovery import (
    MAX_GUIDS_PER_REQUEST,
    MAX_CONCURRENT_REQUESTS,
    BASE_CSV_PARSER_SETTINGS,
    read_mds_into_cache,
    get_discovery_metadata,
    sanitize_tsv_row,
)

REQUIRED_OBJECT_FIELDS = {"dataset_guid", "guid", "display_name"}
OPTIONAL_OBJECT_FIELDS = {"description", "type"}

logging = get_logger("__name__")


async def output_discovery_objects(
    auth,
    dataset_guids=None,
    endpoint=None,
    limit=500,
    use_agg_mds=False,
    guid_type="discovery_metadata",
    only_object_guids=False,
    template=False,
    output_format="tsv",
    output_filename_suffix="",
):
    """
    fetch discovery objects data from a commons and output to {commons}-discovery_objects.tsv or {commons}-discovery_objects.json

    Args:
        auth (Gen3Auth): a Gen3Auth object
        dataset_guids (list of str): a list of datasets to read objects from
        endpoint (str): HOSTNAME of a Gen3 environment, defaults to None
        limit (int): max number of records in one operation, defaults to 500
        use_agg_mds (bool): whether to use AggMDS during export, defaults to False
        guid_type (str): intended GUID type for query, defaults to discovery_metadata
        only_objects_guids (bool): whether to output guids to command line such that they can be piped to another command
        template (bool): whether to output a file with just required headers, to be filled out and then published
        output_format (str): format of output file (can only be either tsv or json), defaults to tsv
        output_filename_suffix (str): additional suffix for the output file name, defaults to ""
    """
    if template:
        output_filename = _create_discovery_objects_filename(auth, "TEMPLATE", ".tsv")
        object_fields = list(REQUIRED_OBJECT_FIELDS)
        object_fields.remove("dataset_guid")
        object_fields.remove("guid")
        object_fields = (
            ["dataset_guid", "guid"]
            + object_fields
            + sorted(list(OPTIONAL_OBJECT_FIELDS))
        )
        with open(output_filename, "w+", encoding="utf-8") as output_file:
            writer = csv.DictWriter(
                output_file,
                **{**BASE_CSV_PARSER_SETTINGS, "fieldnames": object_fields},
            )
            writer.writeheader()
        return output_filename

    if output_format.lower() != "tsv" and output_format != "json":
        logging.error(
            f"Unsupported output file format {output_format}! Only tsv or json is allowed"
        )
        raise ValueError(f"Unsupported output file format {output_format}")

    if endpoint:
        mds = Gen3Metadata(
            auth_provider=auth,
            endpoint=endpoint,
            service_location="mds/aggregate" if use_agg_mds else "mds",
        )
    else:
        mds = Gen3Metadata(
            auth_provider=auth,
            service_location="mds/aggregate" if use_agg_mds else "mds",
        )

    with tempfile.TemporaryDirectory() as metadata_cache_dir:
        partial_metadata, all_fields, num_tags = read_mds_into_cache(
            limit,
            MAX_GUIDS_PER_REQUEST,
            mds,
            guid_type,
            use_agg_mds,
            metadata_cache_dir,
        )

        # output to command line
        if only_object_guids:
            return_objects = []
            for guid in sorted(os.listdir(metadata_cache_dir)):
                if (not dataset_guids) or (guid in dataset_guids):
                    with open(f"{metadata_cache_dir}/{guid}", encoding="utf-8") as f:
                        fetched_metadata = json.load(f)
                        return_objects += (
                            fetched_metadata["objects"]
                            if "objects" in fetched_metadata.keys()
                            else []
                        )
            return return_objects

        # output as TSV
        elif output_format == "tsv":
            output_filename = _create_discovery_objects_filename(
                auth, output_filename_suffix, ".tsv"
            )
            object_fields = REQUIRED_OBJECT_FIELDS.copy()
            all_objects = []

            for guid in sorted(os.listdir(metadata_cache_dir)):
                if (not dataset_guids) or (guid in dataset_guids):
                    with open(f"{metadata_cache_dir}/{guid}", encoding="utf-8") as f:
                        fetched_metadata = json.load(f)
                        curr_objects = (
                            fetched_metadata["objects"]
                            if ("objects" in fetched_metadata.keys())
                            else []
                        )
                        for obj in curr_objects:
                            object_fields |= set(obj.keys())
                            obj["dataset_guid"] = guid
                        all_objects += curr_objects
            object_fields.remove("guid")
            object_fields.remove("dataset_guid")
            object_fields.remove("display_name")
            object_fields = sorted(object_fields)
            object_fields = ["dataset_guid", "guid", "display_name"] + object_fields

            with open(output_filename, "w+", encoding="utf-8") as output_file:
                writer = csv.DictWriter(
                    output_file,
                    **{**BASE_CSV_PARSER_SETTINGS, "fieldnames": object_fields},
                )
                writer.writeheader()
                for obj in all_objects:
                    output_metadata = sanitize_tsv_row(obj)
                    writer.writerow(output_metadata)

        else:
            # output as JSON
            output_filename = _create_discovery_objects_filename(
                auth, output_filename_suffix, ".json"
            )
            output_metadata = []
            dataset_guids = dataset_guids if dataset_guids else partial_metadata.keys()
            for guid in dataset_guids:
                true_guid = guid
                if use_agg_mds:
                    true_guid = guid.split("__")[1]
                metadata = (
                    partial_metadata[guid]["gen3_discovery"]["objects"]
                    if "objects" in partial_metadata[guid]["gen3_discovery"].keys()
                    else []
                )
                for obj in metadata:
                    output_metadata.append({"dataset_guid": true_guid, **obj})

            with open(output_filename, "w+", encoding="utf-8") as output_file:
                output_file.write(json.dumps(output_metadata, indent=4))

        return output_filename


async def publish_discovery_object_metadata(
    auth,
    metadata_filename,
    endpoint=None,
    guid_type="discovery_metadata",
    overwrite=False,
):
    """
    Publish discovery objects from a TSV file

    Args:
        auth (Gen3Auth): a Gen3Auth object
        metadata_filename (str): the file path of the local objects file to be published, must be in TSV format
        endpoint (str): HOSTNAME of a Gen3 environment, defaults to None
        guid_type (str): intended GUID type for publishing, defaults to discovery_metadata
        overwrite (bool): whether to allow replacing objects to a dataset_guid instead of appending
    """
    if not is_valid_object_manifest(metadata_filename):
        raise ValueError(f"Invalid objects file supplied {metadata_filename}")
    if endpoint:
        mds = Gen3Metadata(auth_provider=auth, endpoint=endpoint)
    else:
        mds = Gen3Metadata(auth_provider=auth)

    if not metadata_filename.endswith(".tsv"):
        logging.error(
            f"Unsupported file type supplied {metadata_filename}! Only TSV are allowed."
        )
        raise ValueError(f"Unsupported file type supplied {metadata_filename}")

    delimiter = "\t"
    with open(metadata_filename, encoding="utf-8") as metadata_file:
        csv_parser_setting = {**BASE_CSV_PARSER_SETTINGS, "delimiter": delimiter}
        metadata_reader = csv.DictReader(metadata_file, **{**csv_parser_setting})
        pending_requests = []
        dataset_dict = {}

        for obj_line in metadata_reader:
            # if required fields are missing, display error
            required_fields = list(REQUIRED_OBJECT_FIELDS)
            missing_fields = []
            for field in required_fields:
                if not obj_line[field]:
                    missing_fields.append(field)
            if missing_fields:
                logging.error(f"Missing required field(s) {missing_fields}.")
                raise ValueError(f"Required field(s) missing: {missing_fields}")

            dataset_guid = obj_line.pop("dataset_guid")
            if dataset_guid not in dataset_dict:
                dataset_dict[dataset_guid] = {"objects": []}
            dataset_dict[dataset_guid]["objects"].append(obj_line)

        for dataset_guid in dataset_dict.keys():
            # if dataset_guid already exists, update (noting the use of --overwrite), if it doesn’t already exist, create it
            try:
                curr_dataset_metadata = mds.get(dataset_guid)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404:
                    curr_dataset_metadata = get_discovery_metadata(
                        provided_metadata={}, guid_type=guid_type
                    )
                else:
                    raise

            # allow replacing instead of appending
            if overwrite or not (
                "objects" in curr_dataset_metadata["gen3_discovery"].keys()
            ):
                curr_dataset_metadata["gen3_discovery"]["objects"] = dataset_dict[
                    dataset_guid
                ]["objects"]
            else:
                curr_dataset_dict = {
                    curr_obj["guid"]: curr_obj
                    for curr_obj in curr_dataset_metadata["gen3_discovery"]["objects"]
                }
                for new_obj in dataset_dict[dataset_guid]["objects"]:
                    curr_dataset_dict[new_obj["guid"]] = new_obj

                curr_dataset_metadata["gen3_discovery"]["objects"] = list(
                    curr_dataset_dict.values()
                )

            pending_requests += [
                mds.async_create(dataset_guid, curr_dataset_metadata, overwrite=True)
            ]
            if len(pending_requests) == MAX_CONCURRENT_REQUESTS:
                await asyncio.gather(*pending_requests)
                pending_requests = []
            logging.info(f"Updated objects for Discovery Dataset: {dataset_guid}")
        await asyncio.gather(*pending_requests)


def try_delete_discovery_objects_from_dict(auth, delete_objs):
    """
    Delete discovery objects from a dictionary with schema {<dataset_guid>: [<object_guid1>, <object_guid2>, ...]}

    Args:
        auth (Gen3Auth): a Gen3Auth object
        delete_objs (dict): a dict of dataset_guid keys with a list of object_guid values
    """
    mds = Gen3Metadata(auth_provider=auth)
    for dataset_guid, objects_to_delete in delete_objs.items():
        try:
            metadata = mds.get(dataset_guid)
            if "objects" in metadata["gen3_discovery"].keys():
                curr_objects = metadata["gen3_discovery"]["objects"]
                curr_objects = [
                    obj
                    for obj in curr_objects
                    if (obj["guid"] not in objects_to_delete)
                ]
                metadata["gen3_discovery"]["objects"] = curr_objects
                mds.create(dataset_guid, metadata, overwrite=True)
                logging.info(f"Deleted objects for Discovery Dataset: {dataset_guid}")
        except requests.exceptions.HTTPError as e:
            logging.warning(e)


def try_delete_discovery_objects(auth, guids):
    """
    Delete all discovery objects from one or more datasets

    Args:
        auth (Gen3Auth): a Gen3Auth object
        guids (tuple of str): a tuple of datasets
    """
    mds = Gen3Metadata(auth_provider=auth)
    for arg in guids:
        # delete all objects from dataset_guid
        try:
            metadata = mds.get(arg)
            if metadata["_guid_type"] == "discovery_metadata":
                if "objects" in metadata["gen3_discovery"].keys():
                    metadata["gen3_discovery"]["objects"] = []
                    mds.create(arg, metadata, overwrite=True)
                    logging.info(f"Deleted objects for Discovery Dataset: {arg}")
            else:
                logging.warning(f"{guid} is not discovery metadata. Skipping.")
        except requests.exceptions.HTTPError as e:
            logging.warning(e)


def is_valid_object_manifest(filename, suppress_logs=False):
    """
    Checks a file for required object fields, used for publishing or deleting discovery objects

    Args:
        filename (str): the file to perform validation on
    """
    delimiter = "\t"
    with open(filename, encoding="utf-8") as object_manifest:
        csv_parser_setting = {**BASE_CSV_PARSER_SETTINGS, "delimiter": delimiter}
        objects_reader = csv.DictReader(object_manifest, **{**csv_parser_setting})
        columns = set(objects_reader.fieldnames)
        required_fields = REQUIRED_OBJECT_FIELDS.copy()
        missing_columns = required_fields - columns
        if missing_columns and not suppress_logs:
            logging.error(f"Missing required column(s) {missing_columns}.")
    return REQUIRED_OBJECT_FIELDS.issubset(columns)


def _create_discovery_objects_filename(auth, suffix="", file_extension=".tsv"):
    return (
        "-".join(urlparse(auth.endpoint).netloc.split("."))
        + "-discovery_objects"
        + (f"-{suffix}" if suffix else "")
        + file_extension
    )



import csv
import copy
import json
from cdislogging import get_logger
import tempfile
import asyncio
import os
from urllib.parse import urlparse

import requests.exceptions

from gen3.metadata import Gen3Metadata

from gen3.tools.merge import (
    merge_guids_into_metadata,
    manifest_row_parsers,
    manifests_mapping_config,
    get_guids_for_manifest_row_partial_match,
    get_delimiter_from_extension,
)

from gen3.utils import make_folders_for_filename

MAX_GUIDS_PER_REQUEST = 2000
MAX_CONCURRENT_REQUESTS = 5
BASE_CSV_PARSER_SETTINGS = {
    "delimiter": "\t",
    "quotechar": "",
    "quoting": csv.QUOTE_NONE,
    "escapechar": "\\",
}
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

logging = get_logger("__name__")


async def output_expanded_discovery_metadata(
    auth,
    endpoint=None,
    limit=500,
    use_agg_mds=False,
    guid_type="discovery_metadata",
    output_format="tsv",
    output_filename_suffix="",
):
    """
    fetch discovery metadata from a commons and output to {commons}-{guid_type}.tsv or {commons}-{guid_type}.json

    Args:
        auth (Gen3Auth): a Gen3Auth object
        endpoint (str): HOSTNAME of a Gen3 environment, defaults to None
        limit (int): max number of records in one operation, defaults to 500
        use_agg_mds (bool): whether to use AggMDS during export, defaults to False
        guid_type (str): intended GUID type for query, defaults to discovery_metadata
        output_format (str): format of output file (can only be either tsv or json), defaults to tsv
        output_filename_suffix (str): additional suffix for the output file name, defaults to ""
    """

    if output_format != "tsv" and output_format != "json":
        logging.error(
            f"Unsupported output file format {output_format}! Only tsv or json is allowed"
        )
        raise ValueError(f"Unsupported output file format {output_format}")

    if endpoint:
        mds = Gen3Metadata(
            auth_provider=auth,
            endpoint=endpoint,
            service_location="mds/aggregate" if use_agg_mds else "mds",
        )
    else:
        mds = Gen3Metadata(
            auth_provider=auth,
            service_location="mds/aggregate" if use_agg_mds else "mds",
        )

    with tempfile.TemporaryDirectory() as metadata_cache_dir:
        partial_metadata, all_fields, num_tags = read_mds_into_cache(
            limit,
            MAX_GUIDS_PER_REQUEST,
            mds,
            guid_type,
            use_agg_mds,
            metadata_cache_dir,
        )

        # output as TSV
        if output_format == "tsv":
            output_filename = _create_metadata_output_filename(
                auth, guid_type, output_filename_suffix, ".tsv"
            )
            output_columns = (
                ["guid"]
                # "tags" is flattened to _tag_0 through _tag_n
                + sorted(list(all_fields - set(["tags"])))
                + [f"_tag_{n}" for n in range(num_tags)]
            )
            base_schema = {column: "" for column in output_columns}
            with open(output_filename, "w+", encoding="utf-8") as output_file:
                writer = csv.DictWriter(
                    output_file,
                    **{**BASE_CSV_PARSER_SETTINGS, "fieldnames": output_columns},
                )
                writer.writeheader()

                for guid in sorted(os.listdir(metadata_cache_dir)):
                    with open(f"{metadata_cache_dir}/{guid}", encoding="utf-8") as f:
                        fetched_metadata = json.load(f)
                        flattened_tags = {
                            f"_tag_{tag_num}": f"{tag['category']}: {tag['name']}"
                            for tag_num, tag in enumerate(
                                fetched_metadata.pop("tags", [])
                            )
                        }

                        true_guid = guid
                        if use_agg_mds:
                            true_guid = guid.split("__")[1]
                        output_metadata = sanitize_tsv_row(
                            {
                                **base_schema,
                                **fetched_metadata,
                                **flattened_tags,
                                "guid": true_guid,
                            }
                        )
                        writer.writerow(output_metadata)
        else:
            # output as JSON
            output_filename = _create_metadata_output_filename(
                auth, guid_type, output_filename_suffix, ".json"
            )
            output_metadata = []
            for guid, metadata in partial_metadata.items():
                true_guid = guid
                if use_agg_mds:
                    true_guid = guid.split("__")[1]
                output_metadata.append({"guid": true_guid, **metadata})

            with open(output_filename, "w+", encoding="utf-8") as output_file:
                output_file.write(json.dumps(output_metadata, indent=4))

        return output_filename


def read_mds_into_cache(
    limit, max_guids_per_request, mds, guid_type, use_agg_mds, metadata_cache_dir
):
    """
    Queries an mds instance for all metadata of a guid_type, and writes the data for each guid to a file in metadata_cache_dir

    Args:
        limit (int): max number of records in one operation
        max_guids_per_request (int): max number records for one request to mds
        mds (Gen3Metadata): an instance of Gen3Metadata for an endpoint
        guid_type (str): intended GUID type for query, defaults to discovery_metadata
        use_agg_mds (bool): whether to use AggMDS during export, defaults to False
        metadata_cache_dir (TemporaryDirectory): the temporary directory to write the mds query results to
    """
    all_fields = set()
    num_tags = 0

    for offset in range(0, limit, max_guids_per_request):
        partial_metadata = mds.query(
            f"_guid_type={guid_type}",
            return_full_metadata=True,
            limit=min(limit, max_guids_per_request),
            offset=offset,
            use_agg_mds=use_agg_mds,
        )

        # if agg MDS we will flatten the results as they are in "common" : dict format
        # However this can result in duplicates as the aggregate mds is namespaced to
        # handle this, therefore prefix the commons in front of the guid
        if use_agg_mds:
            partial_metadata = {
                f"{c}__{i}": d
                for c, y in partial_metadata.items()
                for x in y
                for i, d in x.items()
            }

        if len(partial_metadata):
            for guid, guid_metadata in partial_metadata.items():
                with open(
                    f"{metadata_cache_dir}/{guid.replace('/', '_')}",
                    "w+",
                    encoding="utf-8",
                ) as cached_guid_file:
                    guid_discovery_metadata = guid_metadata["gen3_discovery"]
                    json.dump(guid_discovery_metadata, cached_guid_file)
                    all_fields |= set(guid_discovery_metadata.keys())
                    num_tags = max(
                        num_tags, len(guid_discovery_metadata.get("tags", []))
                    )
        else:
            break
    return (partial_metadata, all_fields, num_tags)


def combine_discovery_metadata(
    current_discovery_metadata,
    metadata_file_to_combine,
    discovery_column_to_map_on,
    metadata_column_to_map,
    output_filename,
    metadata_prefix=None,
    exact_match=False,
):
    """
    Combine the provided metadata manifest with the existing Discovery metadata
    in a commons.

    Args:
        current_discovery_metadata (str): filename of TSV containing the current
            discovery metadata
        metadata_file_to_combine (str): filename of TSV containing the metadata
            to combine with the discovery metadata
        discovery_column_to_map_on (str): The column in the current discovery
            metadata to use to map on
        metadata_column_to_map (str): The column in the provided metadata file
            to use to map/merge into the current Discovery metadata
        metadata_prefix (str): Prefix to add to the column names in the provided
            metadata file before final output
        output_filename (str): filename to output combined metadata
        exact_match (str): whether or not the content of discovery_column_to_map_on
            is an EXACT match of the content in metadata_column_to_map.
            Setting this to True when applicable improves the runtime of the
            overall combination.
    """

    logging.info(
        f"Combining '{metadata_file_to_combine}' with '{current_discovery_metadata}' "
        f"by mapping "
        f"'{metadata_column_to_map}' column (in '{metadata_file_to_combine}') to "
        f"current discovery "
        f"metadata's '{discovery_column_to_map_on}' column. Prefixing added "
        f"metadata with: '{metadata_prefix}'. Output will be: {output_filename}"
    )

    # First we need to add a prefix if it was provided
    temporary_prefixed_filename = None
    if metadata_prefix:
        temporary_prefixed_filename = (
            CURRENT_DIR.rstrip("/")
            + "/temp_"
            + os.path.basename(metadata_file_to_combine)
        )
        delimiter = get_delimiter_from_extension(metadata_file_to_combine)
        with open(metadata_file_to_combine, encoding="utf-8") as metadata_file:
            reader = csv.DictReader(metadata_file, delimiter=delimiter)

            new_headers = []
            for header in reader.fieldnames:
                new_headers.append(metadata_prefix + header)

            with open(
                temporary_prefixed_filename, "w", encoding="utf-8"
            ) as prefixed_metadata_file:
                writer = csv.DictWriter(
                    prefixed_metadata_file, fieldnames=new_headers, delimiter=delimiter
                )

                # write the new header out
                writer.writeheader()

                for row in reader:
                    writer.writerow(
                        {
                            metadata_prefix + key: value
                            for (key, value) in row.items()
                            if key
                        }
                    )

    # Now we need to associate the provided metadata with a Discovery GUID

    custom_manifests_mapping_config = copy.deepcopy(manifests_mapping_config)
    custom_manifest_row_parsers = copy.deepcopy(manifest_row_parsers)

    # what column to use as the final GUID for metadata (this MUST exist in the
    # indexing file)
    custom_manifests_mapping_config["guid_column_name"] = "guid"

    # what column from the "metadata file" to use for mapping
    custom_manifests_mapping_config["row_column_name"] = (
        metadata_prefix + metadata_column_to_map
    )
    custom_manifests_mapping_config[
        "indexing_manifest_column_name"
    ] = discovery_column_to_map_on

    # by default, the functions for parsing the manifests and rows assumes a 1:1
    # mapping. There is an additional function provided for partial string matching
    # which we can use here.
    if not exact_match:
        custom_manifest_row_parsers[
            "guids_for_manifest_row"
        ] = get_guids_for_manifest_row_partial_match

    temporary_output_filename = (
        CURRENT_DIR.rstrip("/") + "/temp_" + os.path.basename(output_filename)
    )

    if exact_match:
        merge_guids_into_metadata(
            indexing_manifest=current_discovery_metadata,
            metadata_manifest=temporary_prefixed_filename or metadata_file_to_combine,
            output_filename=temporary_output_filename,
            manifests_mapping_config=custom_manifests_mapping_config,
            manifest_row_parsers=manifest_row_parsers,
            include_all_indexing_cols_in_output=True,
        )

        output_filename = make_folders_for_filename(
            output_filename, current_directory=CURRENT_DIR
        )
    else:
        merge_guids_into_metadata(
            indexing_manifest=current_discovery_metadata,
            metadata_manifest=temporary_prefixed_filename or metadata_file_to_combine,
            output_filename=temporary_output_filename,
            manifests_mapping_config=custom_manifests_mapping_config,
            manifest_row_parsers=custom_manifest_row_parsers,
            include_all_indexing_cols_in_output=False,
        )

        # Now the GUID exists in the output (Discovery GUID has been associated
        # to the provided metadata), so we can easily merge that back
        # into the discovery metadata

        custom_manifests_mapping_config["row_column_name"] = "guid"
        custom_manifests_mapping_config["indexing_manifest_column_name"] = "guid"

        merge_guids_into_metadata(
            indexing_manifest=temporary_output_filename,
            metadata_manifest=current_discovery_metadata,
            output_filename=temporary_output_filename,
            manifests_mapping_config=custom_manifests_mapping_config,
            manifest_row_parsers=manifest_row_parsers,
            include_all_indexing_cols_in_output=True,
        )

        output_filename = make_folders_for_filename(
            output_filename, current_directory=CURRENT_DIR
        )

    # remove rows with GUID of None (this means there was no matching
    # metadata to update in Discovery)
    with open(
        temporary_output_filename, "rt", encoding="utf-8-sig"
    ) as input_file, open(output_filename, "w", encoding="utf-8") as output_file:
        delimiter = get_delimiter_from_extension(temporary_output_filename)
        reader = csv.DictReader(input_file, delimiter=delimiter)

        writer = csv.DictWriter(
            output_file,
            **{
                **BASE_CSV_PARSER_SETTINGS,
                "delimiter": delimiter,
                "fieldnames": reader.fieldnames,
                "extrasaction": "ignore",
            },
        )
        writer.writeheader()

        for row in reader:
            if row.get("guid"):
                writer.writerow(row)

    # cleanup temporary files
    for temporary_file in [
        temp_file
        for temp_file in [temporary_output_filename, temporary_prefixed_filename]
        if temp_file
    ]:
        try:
            os.remove(temporary_file)
        except Exception as exc:
            pass

    logging.info(f"Done. Final combined output: {output_filename}")
    return output_filename


async def publish_discovery_metadata(
    auth,
    metadata_filename,
    endpoint=None,
    omit_empty_values=False,
    guid_type="discovery_metadata",
    guid_field=None,
    is_unregistered_metadata=False,
    reset_unregistered_metadata=False,
    update_registered_metadata=True,
):
    """
    Publish discovery metadata from a tsv or json file

    Args:
        auth (Gen3Auth): a Gen3Auth object
        metadata_filename (str): the file path of the local metadata file to be published, must be in either JSON or TSV format
        endpoint (str): HOSTNAME of a Gen3 environment, defaults to None
        omit_empty_values (bool): whether to exclude fields with empty values from the published discovery metadata, defaults to False
        guid_type (str): intended GUID type for publishing, defaults to discovery_metadata
        guid_field (str): specify a field from the metadata that will be used as GUIDs, if not specified, will try to find a field named "guid" from the metadata, if that field doesn't exist in a certain metadata record, that record will be skipped from publishing, defaults to None
        is_unregistered_metadata (bool): (for use by "study registration" feature only) whether to publish metadata as unregistered study metadata, defaults to False
        reset_unregistered_metadata (bool): (for use by "study registration" feature only) whether to reset existing study metadata back to unregistered study metadata if they exists in the local file, defaults to False
        update_registered_metadata (bool): (for use by "study registration" feature only) whether to update existing study metadata with new values if they exists in the local file, defaults to True
    """
    if endpoint:
        mds = Gen3Metadata(auth_provider=auth, endpoint=endpoint)
    else:
        mds = Gen3Metadata(auth_provider=auth)

    if not metadata_filename:
        metadata_filename = _create_metadata_output_filename(auth, guid_type)

    if not (
        metadata_filename.endswith(".csv")
        or metadata_filename.endswith(".tsv")
        or metadata_filename.endswith(".json")
    ):
        logging.error(
            f"Unsupported file type supplied {metadata_filename}! Only CSV/TSV/JSON are allowed."
        )
        raise ValueError(f"Unsupported file type supplied {metadata_filename}")

    is_json_metadata = False
    if metadata_filename.endswith(".json"):
        is_json_metadata = True

    delimiter = "," if metadata_filename.endswith(".csv") else "\t"

    with open(metadata_filename, encoding="utf-8") as metadata_file:
        if is_json_metadata:
            metadata_reader = json.load(metadata_file)
            tag_columns = []
        else:
            csv_parser_setting = {**BASE_CSV_PARSER_SETTINGS, "delimiter": delimiter}
            if is_unregistered_metadata:
                csv_parser_setting["quoting"] = csv.QUOTE_MINIMAL
                csv_parser_setting["quotechar"] = '"'
            metadata_reader = csv.DictReader(metadata_file, **{**csv_parser_setting})
            tag_columns = [
                column for column in metadata_reader.fieldnames if "_tag_" in column
            ]
        pending_requests = []

        registered_metadata_guids = []
        registered_metadata = {}
        if is_unregistered_metadata:
            if not update_registered_metadata:
                registered_metadata_guids = mds.query(
                    f"_guid_type={guid_type}", limit=2000, offset=0
                )
            else:
                registered_metadata = mds.query(
                    f"_guid_type={guid_type}",
                    return_full_metadata=True,
                    limit=2000,
                    offset=0,
                )
                registered_metadata_guids = registered_metadata.keys()

        for metadata_line in metadata_reader:
            discovery_metadata = {}
            extra_metadata = {}
            if is_json_metadata:
                if "gen3_discovery" in metadata_line:
                    # likely to be a JSON dump from the output_expanded_discovery_metadata() function
                    discovery_metadata = metadata_line.pop("gen3_discovery")
                    # remove unneeded fields
                    try:
                        del metadata_line["_guid_type"]
                    except KeyError:
                        pass
                    extra_metadata = metadata_line
                # no 'gen3_discovery' in JSON, treat entire JSON as discovery metadata
                else:
                    discovery_metadata = metadata_line
            else:
                discovery_metadata = {
                    key: _try_parse(value) for key, value in metadata_line.items()
                }

            if guid_field is None:
                guid = discovery_metadata.pop("guid")
            else:
                guid = discovery_metadata[guid_field]

            if not guid:
                logging.warning(
                    f"{metadata_line} has no GUID information and has been skipped."
                )
                continue

            # remove unneeded fields
            if extra_metadata:
                extra_metadata.pop("guid", None)

            # when publishing unregistered metadata, skip those who are already
            # registered if both reset_unregistered_metadata and
            # update_registered_metadata are set to false
            if (
                is_unregistered_metadata
                and str(guid) in registered_metadata_guids
                and not reset_unregistered_metadata
                and not update_registered_metadata
            ):
                continue

            if len(tag_columns):
                # all columns _tag_0 -> _tag_n are pushed to a "tags" column
                coalesced_tags = [
                    {"name": tag_name.strip(), "category": tag_category.strip()}
                    for tag_category, tag_name in [
                        tag.split(":")
                        for tag in map(discovery_metadata.pop, tag_columns)
                        if tag != ""
                    ]
                ]
                discovery_metadata["tags"] = coalesced_tags

            if omit_empty_values:
                discovery_metadata = {
                    key: value
                    for key, value in discovery_metadata.items()
                    if value not in ["", [], {}]
                }

            new_guid_type = guid_type
            if is_unregistered_metadata:
                if reset_unregistered_metadata or (
                    str(guid) not in registered_metadata_guids
                ):
                    # only set GUID type to "unregistered_discovery_metadata"
                    # for unregistered metadata, or reset_unregistered_metadata is set
                    new_guid_type = f"unregistered_{guid_type}"
                elif str(guid) in registered_metadata_guids:
                    if update_registered_metadata:
                        existing_registered_metadata = {}
                        try:
                            existing_registered_metadata = registered_metadata.get(
                                str(guid)
                            ).get("gen3_discovery")
                        except AttributeError:
                            pass
                        discovery_metadata = {
                            **existing_registered_metadata,
                            **discovery_metadata,
                        }
                    else:
                        logging.warning(f"{guid} is not already registered. Skipping.")
                        continue

            metadata = get_discovery_metadata(
                provided_metadata=discovery_metadata, guid_type=new_guid_type
            )
            if extra_metadata:
                metadata = {**metadata, **extra_metadata}

            pending_requests += [mds.async_create(guid, metadata, overwrite=True)]
            if len(pending_requests) == MAX_CONCURRENT_REQUESTS:
                await asyncio.gather(*pending_requests)
                pending_requests = []

        await asyncio.gather(*pending_requests)


def get_discovery_metadata(
    provided_metadata,
    guid_type="discovery_metadata",
):
    """
    Return a metadata block representing Gen3 Discovery Metadata with the provided
    metadata in the relevant block.

    Args:
        provided_metadata (Dict): Metadata fields you want in Gen3 Discovery
        guid_type (str, optional): value to override the default _guid_type
    """
    return {
        "_guid_type": guid_type,
        "gen3_discovery": provided_metadata,
    }


def try_delete_discovery_guid(auth, guid):
    """
    Deletes all discovery metadata under [guid] if it exists
    """
    mds = Gen3Metadata(auth_provider=auth)
    try:
        metadata = mds.get(guid)
        if metadata["_guid_type"] == "discovery_metadata":
            mds.delete(guid)
        else:
            logging.warning(f"{guid} is not discovery metadata. Skipping.")
    except requests.exceptions.HTTPError as e:
        logging.warning(e)


def sanitize_tsv_row(tsv_row, escape_new_lines=True):
    """
    Cleanup dictionary (tsv_row) so that nested lists/dicts are represented
    in valid JSON and if escape_new_lines, then \n are escaped.

    Args:
        tsv_row (Dict): Dictionary representing a TSV row to write
        escape_new_lines (bool, optional): Whether or not to escape \n as \\n

    Returns:
        Dict: Sanitized input
    """
    sanitized = {}
    for k, v in tsv_row.items():
        if type(v) in [list, dict]:
            sanitized[k] = json.dumps(v)
        elif (type(v) == str) and escape_new_lines:
            sanitized[k] = v.replace("\n", "\\n")
        elif type(v) == int:
            sanitized[k] = v
    return sanitized


def _try_parse(data):
    if data:
        data = data.replace("\\n", "\n")
        try:
            return json.loads(data)
        except json.JSONDecodeError:
            return data
    return ""


def _create_metadata_output_filename(auth, guid_type, suffix="", file_extension=".tsv"):
    return (
        "-".join(urlparse(auth.endpoint).netloc.split("."))
        + f"-{guid_type}"
        + (f"-{suffix}" if suffix else "")
        + file_extension
    )



import re

from cdislogging import get_logger

logging = get_logger("__name__")

# For more details about this regex, see the function that uses it
DBGAP_ACCESSION_REGEX = (
    "(?P<phsid>phs(?P<phsid_number>[0-9]+))"
    "(.(?P<participant_set>p(?P<participant_set_number>[0-9]+))){0,1}"
    "(.(?P<version>v(?P<version_number>[0-9]+))){0,1}"
    "(.(?P<consent>c(?P<consent_number>[0-9]+)+)){0,1}"
)


def get_dbgap_accession_as_parts(phsid):
    """
    Return a dictionary containing the various parts of the provided
    dbGaP Accession (AKA phsid).

    Uses a regex to match an assession number that has information in forms like:
      phs000123.c1
      phs000123.v3.p1.c3
      phs000123.c3
      phs000123.v3.p4.c1
      phs000123

    This separates out each part of the accession with named groups and includes
    parts that include only the numbered value (which is needed in some NIH APIs)

    A "picture" is worth a 1000 words:

    Example for `phs000123.c1`:
      Named groups
      phsid                   phs000123
      phsid_number            000123
      version                 None
      version_number          None
      participant_set         None
      participant_set_number  None
      consent                 c1
      consent_number          1

    Args:
        phsid (str): The dbGaP Accession (AKA phsid)

    Returns:
        dict[str]: A standardized dictionary (you can always expect these keys)
                   with the values parsed from the provided dbGaP Accession
                   Example if provided `phs000123.c1`: {
                        "phsid": "phs000123",
                        "phsid_number": "000123",
                        "version": "",
                        "version_number": "",
                        "participant_set": "",
                        "participant_set_number": "",
                        "consent": "c1",
                        "consent_number": "1",
                    }

                    NOTE: the "*_number" fields are still represented as strings.
                    NOTE2: the regex groups that return None will be represented
                           as empty strings (for easier upstream str concat-ing)
    """
    access_number_matcher = re.compile(DBGAP_ACCESSION_REGEX)
    raw_phs_match = access_number_matcher.match(phsid)
    phs_match = {}

    if raw_phs_match:
        phs_match = raw_phs_match.groupdict()

    standardized_phs_match = {}
    for key, value in phs_match.items():
        if value is None:
            standardized_phs_match[key] = ""
            continue

        standardized_phs_match[key] = value

    return standardized_phs_match



import click
import json
import sys
from gen3.wss import Gen3WsStorage, wsurl_to_tokens


def clean_path(path):
    """
    Add ws:///@user/ prefix if necessary
    """
    if path[0:3] != "ws:":
        while path and path[0] == "/":
            path = path[1:]
        path = "ws:///@user/" + path
    return path


@click.command()
@click.argument("path", default="")
@click.pass_context
def ls(ctx, path=""):
    """List the given workspace key"""
    clean = clean_path(path)
    auth_provider = ctx.obj["auth_factory"].get()
    wss = Gen3WsStorage(auth_provider)
    print(json.dumps(wss.ls_path(clean)))


@click.command()
@click.argument("path")
@click.pass_context
def rm(ctx, path):
    """Remove the given workspace key"""
    clean = clean_path(path)
    auth_provider = ctx.obj["auth_factory"].get()
    wss = Gen3WsStorage(auth_provider)
    print(json.dumps(wss.rm_path(clean)))


@click.command()
@click.argument("path")
@click.pass_context
def download_url(ctx, path):
    """Download url for the given workspace key"""
    tokens = wsurl_to_tokens(clean_path(path))
    auth_provider = ctx.obj["auth_factory"].get()
    wss = Gen3WsStorage(auth_provider)
    print(json.dumps(wss.download_url(tokens[0], tokens[1])))


@click.command()
@click.argument("path")
@click.pass_context
def upload_url(ctx, path):
    """Upload url for the given workspace key"""
    tokens = wsurl_to_tokens(clean_path(path))
    auth_provider = ctx.obj["auth_factory"].get()
    wss = Gen3WsStorage(auth_provider)
    print(json.dumps(wss.upload_url(tokens[0], tokens[1])))


@click.command()
@click.argument("src")
@click.argument("dest")
@click.pass_context
def copy(ctx, src, dest):
    """Upload url for the given workspace key"""
    auth_provider = ctx.obj["auth_factory"].get()
    wss = Gen3WsStorage(auth_provider)
    wss.copy(src, dest)


@click.group()
def wss():
    """[unfinished] Commands for Workspace Storage Service"""
    pass


wss.add_command(ls, name="ls")
wss.add_command(copy, name="cp")
wss.add_command(download_url, name="download-url")
wss.add_command(upload_url, name="upload-url")
wss.add_command(rm, name="rm")



import os
import subprocess

from cdislogging import get_logger
from gen3.auth import Gen3Auth, Gen3AuthError

logger = get_logger("__name__")


class Gen3Wrap:
    def __init__(self, auth: Gen3Auth, command_args: tuple):
        """
        auth : Gen3Auth instance
        command_args: A tuple consisting of all the commands sent to the `gen3 run` tool
        """
        self.auth = auth
        self.command_args = command_args

    def run_command(self):
        """
        Take the command args and run a subprocess with appropriate access token in the env var
        """
        cmd = list(self.command_args)
        try:
            os.environ["GEN3_TOKEN"] = self.auth.get_access_token()
        except Gen3AuthError as e:
            logger.error(f"ERROR getting Gen3 Access Token:", e)
            raise
        logger.info(
            f"Running the command {self.command_args} with gen3 access token in environment variable"
        )
        try:
            subprocess.run(cmd, stderr=subprocess.STDOUT)
        except Exception as e:
            logger.error(f"ERROR while running '{cmd}':", e)
            raise



import asyncio
import click
import csv
import logging
import os
import sys
from urllib.parse import urlparse

from gen3.tools import indexing
from gen3.tools.indexing.validate_manifest_format import is_valid_manifest_format
from gen3.tools.indexing.index_manifest import (
    index_object_manifest,
    delete_all_guids,
)
from gen3.tools.indexing.verify_manifest import manifest_row_parsers
from gen3.utils import get_or_create_event_loop_for_thread
from gen3.tools.metadata.crosswalk import (
    publish_crosswalk_metadata,
    read_crosswalk_metadata,
)


@click.group()
def objects():
    """Commands for reading and editing objects"""
    pass


@objects.group()
def manifest():
    """For working with minimal object metadata manifests"""
    pass


@objects.group()
def crosswalk():
    """For working with cross-commons subject crosswalk data"""
    pass


@click.command(
    help="Creates an object manifest of current data by reading from the Gen3 instance."
)
@click.option(
    "--output-file",
    "output_file",
    default="object-manifest.csv",
    help="filename for output",
    type=click.Path(writable=True),
    show_default=True,
)
@click.option(
    "--num-processes",
    "num_processes",
    default=4,
    help="Number of parallel Python processes to start",
    type=int,
    show_default=True,
)
@click.option(
    "--max-concurrent-requests",
    "max_concurrent_requests",
    help="Maximum concurrent requests to Gen3",
    default=24,
    type=int,
    show_default=True,
)
@click.option(
    "--input-manifest",
    "input_manifest",
    help="Input file. Read available object data only for records referenced in this file. "
    "Currently requires at a minimum an `m5d` column with checksum.",
    default=None,
    type=click.Path(writable=True),
    show_default=True,
)
@click.option(
    "--python-subprocess-command",
    "python_subprocess_command",
    help="Command used to execute a python process. By default you should not need "
    "to change this, but if you are running something like MacOS and only "
    "installed Python 3.x you may need to specify 'python3'",
    default="python",
    show_default=True,
)
@click.pass_context
def objects_manifest_read(
    ctx,
    output_file,
    num_processes,
    max_concurrent_requests,
    input_manifest,
    python_subprocess_command,
):
    auth = ctx.obj["auth_factory"].get()
    loop = get_or_create_event_loop_for_thread()
    click.echo(f"Getting minimal object metadata from {auth.endpoint}")
    loop.run_until_complete(
        indexing.async_download_object_manifest(
            auth.endpoint,
            output_filename=output_file,
            num_processes=num_processes,
            max_concurrent_requests=max_concurrent_requests,
            input_manifest=input_manifest,
            python_subprocess_command=python_subprocess_command,
        )
    )
    click.echo(output_file)


@click.command(
    help="Verifies that a Gen3 instance contains the data in the specified object manifest."
)
@click.argument("file", required=False)
@click.option(
    "--max-concurrent-requests",
    "max_concurrent_requests",
    help="Maximum concurrent requests to Gen3",
    default=24,
    type=int,
    show_default=True,
)
@click.pass_context
def objects_manifest_verify(ctx, file, max_concurrent_requests):
    auth = ctx.obj["auth_factory"].get()
    loop = get_or_create_event_loop_for_thread()

    if not file:
        file = click.prompt("Enter Discovery metadata file path to publish")

    click.echo(f"Verifying {file}...\n    Against: {auth.endpoint}")

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    loop.run_until_complete(
        indexing.async_verify_object_manifest(
            auth.endpoint,
            manifest_file=file,
            max_concurrent_requests=max_concurrent_requests,
        )
    )


@click.command(
    help="Validates that the format of the specified object manifest is correct."
)
@click.argument("file", required=False)
@click.option(
    "--allowed-protocols",
    "allowed_protocols",
    help="""
    space-delimited string list of allowed protocols for url validation.

    Note that if allowed_protocols is provided, url values will only be
    validated using the provided protocols (e.g. if
    allowed_protocols="http https" an error would be raised when
    validating "s3://valid_bucket/valid_key" url)
    """,
    default="s3 gs",
    show_default=True,
)
@click.option(
    "--allow-base64-encoded-md5",
    "allow_base64_encoded_md5",
    help="""
    whether or not Base64 encoded md5 values are allowed.

    if False, only hexadecimal encoded 128-bit md5 values are considered valid,
    and Base64 encoded values will be logged as errors.

    if arg provided, both hexadecimal and Base64 encoded 128-bit md5 values are considered
    valid
    """,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--error-on-empty-url",
    "error_on_empty_url",
    help="""
    whether to treat completely empty url values as errors

    for the following example manifest, if error_on_empty_url is False,
    a warning would be logged for the completely empty url value on
    line 2. if error_on_empty_url is True, an error would be generated
    instead
    ```
    md5,url,size
    1596f493ba9ec53023fca640fb69bd3b,,42
    ```

    note that regardless of error_on_empty_url, errors will be
    generated no matter what for arrays or quotes from which urls could
    not be extracted. for example, for the following manifest, errors
    would be generated for the url value on lines 2, 3, 4, and 5
    ```
    md5,url,size
    1596f493ba9ec53023fca640fb69bd30,"",42
    1596f493ba9ec53023fca640fb69bd31,'',43
    1596f493ba9ec53023fca640fb69bd32,[],44
    1596f493ba9ec53023fca640fb69bd33,["", ""],45
    ```
    """,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--line-limit",
    "line_limit",
    help="""
    number of lines in manifest to validate including the header. if
    not provided, every line is validated.

    This can be helpful as a way to only attempt validation of a few rows to get
    an idea if there are large format issues without needing to run against every
    row in the manifest.
    """,
    default=None,
    type=int,
    show_default=True,
)
@click.pass_context
def objects_manifest_validate_format(
    ctx,
    file,
    allowed_protocols,
    allow_base64_encoded_md5,
    error_on_empty_url,
    line_limit,
):
    if not file:
        file = click.prompt("Enter Discovery metadata file path to validate format for")

    is_valid = is_valid_manifest_format(
        manifest_path=file,
        column_names_to_enums=None,
        allowed_protocols=allowed_protocols.split(" "),
        allow_base64_encoded_md5=allow_base64_encoded_md5,
        error_on_empty_url=error_on_empty_url,
        line_limit=line_limit,
    )

    # non-zero exit code
    if not is_valid:
        sys.exit(1)


@click.command(help="Publishes specified object manifest to Gen3 instance.")
@click.argument("file", required=False)
@click.option(
    "--thread-num",
    "thread_num",
    help="number of threads for indexing",
    default=8,
    type=int,
    show_default=True,
)
@click.option(
    "--append-urls",
    "append_urls",
    help="""
    If supplied, will append urls for existing records. e.g. existing urls will
    still exist and new ones will be added

    By default the newly provided urls will REPLACE existing urls
    """,
    is_flag=True,
    show_default=True,
)
@click.option(
    "--manifest-file-delimiter",
    "manifest_file_delimiter",
    help="string character that delimites the file (tab or comma). Defaults to tab.",
    default=None,
    show_default=True,
)
@click.option(
    "--out-manifest-file",
    "out_manifest_file",
    help="The path to output manifest",
    default="indexing-output-manifest.csv",
    show_default=True,
)
@click.option(
    "--force-metadata-columns-even-if-empty",
    "force_metadata_columns_even_if_empty",
    help="force the creation of a metadata column entry for a GUID even if the value "
    "is empty. Enabling this will force the creation of metadata entries for every column.",
    is_flag=True,
)
@click.pass_context
def objects_manifest_publish(
    ctx,
    file,
    thread_num,
    append_urls,
    manifest_file_delimiter,
    out_manifest_file,
    force_metadata_columns_even_if_empty,
):
    auth = ctx.obj["auth_factory"].get()
    loop = get_or_create_event_loop_for_thread()

    if not file:
        file = click.prompt("Enter Discovery metadata file path to publish")

    click.echo(
        f"Publishing/writing object data from {file}...\n    to: {auth.endpoint}"
    )

    index_object_manifest(
        commons_url=auth.endpoint,
        manifest_file=file,
        thread_num=thread_num,
        auth=auth,
        replace_urls=not append_urls,
        manifest_file_delimiter=manifest_file_delimiter,
        output_filename=out_manifest_file,
        submit_additional_metadata_columns=True,
        force_metadata_columns_even_if_empty=force_metadata_columns_even_if_empty,
    )


@click.command(help="Deletes specified object manifest to Gen3 instance.")
@click.argument("file", required=True)
@click.pass_context
def objects_manifest_delete_all_guids(ctx, file):
    auth = ctx.obj["auth_factory"].get()
    loop = get_or_create_event_loop_for_thread()

    if not file:
        file = click.prompt("Enter Discovery metadata file path to delete")

    click.echo(f"        DELETING ALL GUIDS\n  from: {file}\n    in: {auth.endpoint}")
    click.confirm(
        f"Are you sure you want to DELETE ALL GUIDS in {auth.endpoint} as specified by this file: {file}?",
        abort=True,
    )
    click.confirm(
        f"Please confirm again, this is irreversible. All GUIDs specified specified in {file} will be deleted from {auth.endpoint}. You are sure?",
        abort=True,
    )

    delete_all_guids(auth, file)


manifest.add_command(objects_manifest_read, name="read")
manifest.add_command(objects_manifest_verify, name="verify")
manifest.add_command(objects_manifest_validate_format, name="validate-manifest-format")
manifest.add_command(objects_manifest_publish, name="publish")
manifest.add_command(objects_manifest_delete_all_guids, name="delete-all-guids")


@click.command(
    help=(
        """
    Publishes specified crosswalk from local files to Gen3 instance and merges
    with existing crosswalk data already in Gen3.\n

    FILE\n
    \tA file path to a CSV containing mapping identifiers. Has a specialized
    \tcolumn naming format. Column names MUST be pipe-delimited and contain:\n
    \t\tcommons url | identifier type | identifier name\n
    
    \tYou can have any number of columns for mapping.\n
    """
    )
)
@click.argument(
    "file",
    type=click.Path(writable=True),
    required=True,
)
@click.option(
    "-m",
    "--mapping_methodology",
    "mapping_methodologies",
    required=True,
    multiple=True,
    help="A string description of how the mapping in the provided file was obtained.",
)
@click.option(
    "--info",
    "info_file",
    type=click.Path(writable=True),
    required=False,
    help=(
        "File with an additional description about the identifiers in the crosswalk file. "
        "The format is a CSV with columns: commons url,identifier name,description. Be sure "
        "to use the same commons url and identifier name used in the crosswalk file."
    ),
)
@click.pass_context
def objects_crosswalk_publish(ctx, file, info_file, mapping_methodologies):
    auth = ctx.obj["auth_factory"].get()

    loop = get_or_create_event_loop_for_thread()
    loop.run_until_complete(
        publish_crosswalk_metadata(
            auth=auth,
            file=file,
            info_file=info_file,
            mapping_methodologies=mapping_methodologies,
        )
    )


@click.command(help="Reads crosswalk data from Gen3 instance into local files.")
@click.option(
    "--output-file",
    "output_file",
    default="crosswalk.csv",
    help="filename for output",
    type=click.Path(writable=True),
    show_default=True,
)
@click.pass_context
def objects_crosswalk_read(ctx, output_file):
    auth = ctx.obj["auth_factory"].get()

    loop = get_or_create_event_loop_for_thread()
    loop.run_until_complete(
        read_crosswalk_metadata(auth=auth, output_filename=output_file)
    )


@click.command(
    help="Verifies specified crosswalk data in local files exists in Gen3 instance."
)
@click.pass_context
def objects_crosswalk_verify(ctx):
    raise NotImplementedError("`gen3 objects crosswalk verify` is not implemented yet.")


@click.command(
    help="Deletes specified crosswalk data in local files from Gen3 instance."
)
@click.pass_context
def objects_crosswalk_delete(ctx):
    raise NotImplementedError("`gen3 objects crosswalk delete` is not implemented yet.")


crosswalk.add_command(objects_crosswalk_publish, name="publish")
crosswalk.add_command(objects_crosswalk_read, name="read")
crosswalk.add_command(objects_crosswalk_verify, name="verify")
crosswalk.add_command(objects_crosswalk_delete, name="delete")



class ExternalMetadataSourceInterface(object):
    """
    A simple interface for external metadata sources. The idea is to have
    consistency that allows combining the outputs and making similar function
    calls with similar input.

    test0 = dbgapFHIR()
    metadata_0 = test0.get_metadata_for_ids(["foo", "bar"])

    test1 = NIHReporter()
    metadata_1 = test1.get_metadata_for_ids(["foo", "bar"])

    test2 = dbgapDOI()
    metadata_2 = test2.get_metadata_for_ids(["foo", "bar"])

    test3 = somethingElse()
    metadata_3 = test3.get_metadata_for_ids(["foo", "bar"])

    all_metadata = metadata_0 + metadata_1 + metadata_2 + metadata_3
    """

    def __init__(
        self,
        api="",
        auth_provider=None,
    ):
        self.api = api
        self._auth_provider = auth_provider

    def get_metadata_for_ids(ids):
        """
        Returns a dictionary with the id as the key and associated metadata
        as another dictionary of values.

        Example, given: ["foo", "bar"], return:
        {
            "foo": {"name": "Foo", "description": "this is something"},
            "bar": {"name": "Bar", "description": "this is also something"},
        }

        Args:
            ids (List[str]): list of IDs to query for

        Returns:
            Dict[dict]: metadata for each of the provided IDs (which
                        are the keys in the returned dict)
        """
        raise NotImplementedError()



import backoff
import collections
import copy
import csv
import time

from cdislogging import get_logger
import fhirclient.models.researchstudy as researchstudy
from fhirclient import client
from fhirclient.models.fhirreference import FHIRReference
from fhirclient.models.identifier import Identifier
from fhirclient.models.meta import Meta

from gen3.external import ExternalMetadataSourceInterface
from gen3.external.nih.utils import get_dbgap_accession_as_parts
from gen3.tools.metadata.discovery import sanitize_tsv_row, BASE_CSV_PARSER_SETTINGS
from gen3.utils import DEFAULT_BACKOFF_SETTINGS

logging = get_logger("__name__")


class dbgapFHIR(ExternalMetadataSourceInterface):
    """
    Class for interacting with National Center for
    Biotechnology Information (NCBI)'s database of Genotypes and Phenotypes (dbGaP)
    FHIR API

    Example Usage:

        ```
        from gen3.external.nih.dbgap_fhir import dbgapFHIR


        def main():
            studies = [
                "phs000007",
                "phs000166.c3",
                "phs000179.v1.p1.c1",
            ]
            dbgapfhir = dbgapFHIR()
            simplified_data = dbgapfhir.get_metadata_for_ids(ids=studies)

            file_name = "data_file.tsv"
            dbgapFHIR.write_data_to_file(simplified_data, file_name)


        if __name__ == "__main__":
            main()
        ```

    This provides utilities for getting public study-level data using a third-party
    FHIR client library behind the scenes. This also provides some opinionated methods
    for transforming the raw dbGaP data into a "simplified" format.

    dbGaP's FHIR naming conventions and organization doesn't translate well to simple
    key-value pairs. We want to translate the info provided to simplify its understanding
    and allow ingest into our Metadata API as denormalized as possible, in other words, reduce
    the verbosity and nesting.

    This code attempts to remain the least "hard-coded" as possible, prefering to
    add special cases for known deviations of structure rather than writing code to literally
    parse every field we want.

    Note that either way, if dbGaP changes their structure or we want additional information,
    some of these functions may need to get reworked (or calling functions will need to do
    "hard-coded" parsing in addition to using the raw FHIR output objects / FHIR library).

    Known dbGaP FHIR fields interpretted as Lists:
        identifier
        keyword
        Citers
        condition
        StudyConsents
        partOf

    Known dbGaP FHIR fields interpretted as Strings:
        description
        id
        status
        resourceType
        title
        sponsor
        StudyOverviewUrl
        ReleaseDate
        focus
        enrollment
        category

    Known dbGaP FHIR fields interpretted as Strings but generally represent Integers:
        NumMolecularDatasets
        NumSubjects
        NumSubStudies
        NumPhenotypeDatasets
        NumSamples
        NumAnalyses
        NumVariables
        NumDocuments

    Attributes:
        api (str): URL to dbGaP's FHIR API
        DBGAP_SLEEP_TIME_TO_AVOID_RATE_LIMITS (float): Time to sleep in seconds between
            requests to dbGaP's FHIR server. This was tuned to avoid rate limiting
            experienced with smaller values.

            TODO: implement exponential backoff using backoff library used elsewhere
        fhir_client (fhirclient.client.FHIRClient): client class for interacting with FHIR
        SUSPECTED_SINGLE_ITEM_LIST_FIELDS (List[str]): dbGaP FHIR fields that are lists
            in the API, but in all testing only yield (and logicially should only yield)
            a single result. The single item in the list will be pulled out so that the
            final field is NOT a list.
        UNNECESSARY_FIELDS(List[str]): fields to remove in final result
    """

    # this is necessary in addition to exponential backoff to reduce the number
    # of forced backoff requests (too many requests in a timeframe seem to trip
    # some rate limiting on NIH's side)
    DBGAP_SLEEP_TIME_TO_AVOID_RATE_LIMITS = 0.6

    SUSPECTED_SINGLE_ITEM_LIST_FIELDS = [
        "sponsor",
        "StudyOverviewUrl",
        "ReleaseDate",
        "focus",
        "enrollment",
        "category",
    ]

    UNNECESSARY_FIELDS = [
        "meta",
        "StudyMarkersets",
        "security",
        "ComputedAncestry",
        "MolecularDataTypes",
    ]

    DISCLAIMER = (
        "This information was retrieved from dbGaP's FHIR API for "
        "discoverability purposes and may not contain fully up-to-date "
        "information. Please refer to the official dbGaP FHIR server for up-to-date "
        "FHIR data."
    )

    def __init__(
        self,
        api="https://dbgap-api.ncbi.nlm.nih.gov/fhir/x1",
        auth_provider=None,
        fhir_app_id="gen3_sdk",
        suspected_single_item_list_fields=None,
        unnecessary_fields=None,
        disclaimer=None,
    ):
        """
        Inializate an instance of the class

        Args:
            api (str, optional): URL to dbGaP's FHIR API
            auth_provider (None, optional): Description
            fhir_app_id (str, optional): name of app_id to pass into the FHIR client library
        """
        super(dbgapFHIR, self).__init__(api, auth_provider)

        fhir_client_settings = {
            "app_id": fhir_app_id,
            "api_base": api,
        }
        self.fhir_client = client.FHIRClient(settings=fhir_client_settings)
        self.suspected_single_item_list_fields = (
            suspected_single_item_list_fields
            or dbgapFHIR.SUSPECTED_SINGLE_ITEM_LIST_FIELDS
        )
        self.unnecessary_fields = unnecessary_fields or dbgapFHIR.UNNECESSARY_FIELDS
        self.disclaimer = disclaimer or dbgapFHIR.DISCLAIMER

    def get_metadata_for_ids(self, ids):
        """
        Return simplified FHIR metadata for each of the provided IDs.

        NOTE: This is NOT the raw metadata from the dbGaP FHIR server. This is a
              denormalized and simplified representation of that data intended
              to both reduce the nesting and ease the understanding of the
              information.

        Args:
            ids (List[str]): list of IDs to query FHIR for

        Returns:
            Dict[dict]: metadata for each of the provided IDs (which
                        are the keys in the returned dict)
        """
        logging.info("Getting dbGaP FHIR metadata...")
        logging.debug("Provided ids for dbGaP FHIR metadata: {ids}")

        simplified_data = {}

        original_phsid_to_standardized = {}

        standardized_phsids = []
        for phsid in ids:
            phsid_parts = get_dbgap_accession_as_parts(phsid)

            # allow form of ONLY "phs000123", dbGaP FHIR API cannot handle
            # isolating a specific version
            standardized_phsid = phsid_parts["phsid"]

            # we want the final returned value to include whatever format was
            # used in the original call, so even though we are using the standardized
            # phs against the API, we need to convert BACK to the original
            # before returning. This dict maintains that mapping.
            original_phsid_to_standardized[phsid] = standardized_phsid

            standardized_phsids.append(standardized_phsid)

        all_data = self.get_dbgap_fhir_objects_for_studies(standardized_phsids)

        standardized_phsids_data = {}
        for phsid in ids:
            standardized_phsid = original_phsid_to_standardized[phsid]

            # it's possible we've already simplified the data for this phsid
            # (ex: phs00007.p1.v1.c1 AND phs00007.p1.v1.c2 rely on data for the
            #      overall study, phs00007. So there's no need to parse this twice )
            #
            # NOTE: This is because dbGaP has metadata at the STUDY level, not
            #       consent group level (which is usually how we organize
            #       datasets in Gen3)
            if standardized_phsid in standardized_phsids_data:
                data = standardized_phsids_data[standardized_phsid]
            else:
                # error should be logged by `get_dbgap_fhir_objects_for_studies`
                # so just skip here
                if not standardized_phsid in all_data:
                    continue

                data = self.get_simplified_data_from_object(
                    all_data[standardized_phsid]
                )

            # custom fields
            data["ResearchStudyURL"] = (
                self.api.rstrip("/") + "/ResearchStudy/" + standardized_phsid
            )
            data["Disclaimer"] = self.disclaimer

            logging.debug(f"simplified {phsid} study data: {data}")
            standardized_phsids_data[standardized_phsid] = data
            simplified_data[phsid] = data

        return simplified_data

    def get_dbgap_fhir_objects_for_studies(self, ids):
        """
        Return FHIR Objects

        Args:
            ids (List[str]): list of IDs to query FHIR for

        Returns:
            Dict[fhirclient.models.ResearchStudy]: FHIR objects for each of the
                provided IDs (which are the keys in the returned dict)
        """
        all_data = {}
        for study_id in ids:
            try:
                study = self._get_dbgap_fhir_objects_for_study(study_id)
            except Exception as exc:
                logging.error(f"unable to get {study_id}, skipping. Error: {exc}")
                continue

            # dbGaP API rate limits if we go too fast
            # with "429 Client Error: Too Many Requests for url"
            if dbgapFHIR.DBGAP_SLEEP_TIME_TO_AVOID_RATE_LIMITS > 0:
                logging.debug(
                    f"sleeping for {dbgapFHIR.DBGAP_SLEEP_TIME_TO_AVOID_RATE_LIMITS} seconds so we don't make dbGaP's API mad."
                )
                time.sleep(dbgapFHIR.DBGAP_SLEEP_TIME_TO_AVOID_RATE_LIMITS)

            logging.debug(f"raw study data: {study.as_json()}")
            all_data[study_id] = study

        return all_data

    @backoff.on_exception(
        backoff.expo,
        Exception,
        max_tries=7,
        **{
            key: value
            for key, value in DEFAULT_BACKOFF_SETTINGS.items()
            if key != "max_tries"
        },
    )
    def _get_dbgap_fhir_objects_for_study(self, phsid):
        logging.info(f"Getting {phsid} from dbGaP FHIR API...")
        study = None

        try:
            study = researchstudy.ResearchStudy.read(phsid, self.fhir_client.server)
        except Exception as exc:
            logging.warning(f"Unable to get {phsid}, skipping. Error: {exc}")
            raise

        return study

    @staticmethod
    def write_data_to_file(all_data, file_name):
        """
        Write the values of every key in the provided dictionary to a tabular file (TSV).

        This will convert any top level fields that are Dicts to a JSON representation
        such that a single all_data[key] becomes a single row in the outputed TSV.

        Args:
            all_data (Dict[Dict]): study metadata
            file_name (str): output file name
        """
        logging.info(f"Beginning to write data to file: {file_name}")
        data_file = open(file_name, "w")

        headers = set()

        for study_id, study_data in all_data.items():
            for item in study_data.keys():
                headers.add(item)

        csv_writer = csv.DictWriter(
            data_file,
            **{
                **BASE_CSV_PARSER_SETTINGS,
                "fieldnames": sorted(headers),
                "extrasaction": "ignore",
            },
        )
        csv_writer.writeheader()

        for study_id, study_data in all_data.items():
            # Writing data of CSV file
            csv_writer.writerow(sanitize_tsv_row(study_data))

        logging.info(f"Done writing data to file: {file_name}")

    def get_simplified_data_from_object(self, fhir_object):
        """
        An attempt at simplifying the representation of dbGaP's FHIR data to something
        more easily readable and more flat.

        Args:
            fhir_object (fhirclient.models.*): Any FHIR object

        Returns:
            dict: dictionary of all simplified data

        Raises:
            Exception: If an extension has more than 1 value attributes.
        """
        all_data = {}

        # special handling for FHIR "extensions", which are terms in the response
        # that are not part of the standard FHIR model (in other words, they are
        # custom fields that dbGaP has added)
        for extension in getattr(fhir_object, "extension", None) or []:
            for potential_value_type in dir(extension):
                if potential_value_type == "extension" and extension.extension:
                    ext_name = extension.url.split("/")[-1].split("-")[-1]
                    if not all_data.get(ext_name):
                        all_data[ext_name] = []

                    all_data[ext_name].append(
                        self.get_simplified_data_from_object(extension)
                    )

            for potential_value_type in dir(extension):
                found_value = False
                if potential_value_type.startswith("value") and getattr(
                    extension, potential_value_type, None
                ):
                    if found_value:
                        raise Exception(
                            f"Found 2 value attributes for extension: {extension.url}, can only "
                            f"handle one. First: {all_data[extension.url]} "
                            f"Second: {potential_value_type}: {getattr(extension, potential_value_type, None)}"
                        )
                    value = getattr(extension, potential_value_type, None)
                    # it's possible the value is another FHIR class, so try to represent it as json
                    try:
                        value = value.as_json()
                        if value.get("value"):
                            value = value["value"]
                        elif value.get("display"):
                            value = value["display"]
                        elif value.get("code"):
                            value = value["code"]
                    except AttributeError:
                        # assume we're dealing with a default Python type and continue
                        pass

                    ext_name = extension.url.split("/")[-1].split("-")[-1]
                    # if it's a Mapping, we weren't able to simplify the value,
                    # so just leave the existing entry in all_data
                    if not isinstance(value, collections.abc.Mapping):
                        if not all_data.get(ext_name):
                            all_data[ext_name] = []

                        all_data[ext_name].append(value)

                    found_value = True

        # use libraries built-in parsing for standard FHIR fields
        non_extension_data = fhir_object.as_json()

        # The library attempts to parse extensions, but we already handled
        # the dbGaP extension data above, so remove the raw info from the library
        # and replace with our parsing
        del non_extension_data["extension"]
        all_data.update(non_extension_data)

        # simplify and clean up everything
        simplified_data = dbgapFHIR._get_simplified_data(fhir_object)
        all_data.update(simplified_data)
        self._flatten_relevant_fields(all_data)
        self._remove_unecessary_fields(all_data)
        self._capitalize_top_level_keys(all_data)

        return all_data

    @staticmethod
    def _get_simplified_data(fhir_object):
        simplified_data = {}

        variables_to_simplify = set(
            [
                "category",
                "condition",
                "keyword",
                "focus",
                "identifier",
                "security",
                "sponsor",
                "security",
            ]
        )

        # per docs, .elementProperties() returns:
        #       Returns a list of tuples, one tuple for each property that should be
        #       serialized, as: ("name", "json_name", type, is_list, "of_many", not_optional)
        for item in fhir_object.elementProperties():
            variables_to_simplify.add(item[0])

        for var in variables_to_simplify:
            simplified_var = dbgapFHIR._get_simple_text_from_fhir_object(
                fhir_object=fhir_object, list_variable=var
            )
            if simplified_var:
                simplified_data[var] = simplified_var

        return simplified_data

    @staticmethod
    def _get_simple_text_from_fhir_object(fhir_object, list_variable):
        output = []
        if type(getattr(fhir_object, list_variable, None)) is list:
            for item in getattr(fhir_object, list_variable, None) or []:
                simple_representation = getattr(item, "display", {})
                simple_representation = simple_representation or getattr(
                    item, "value", {}
                )
                simple_representation = simple_representation or getattr(
                    item, "text", {}
                )
                simple_representation = simple_representation or getattr(
                    item, "reference", {}
                )
                simple_representation = (
                    simple_representation
                    or dbgapFHIR._get_simple_text_from_fhir_object(
                        getattr(item, "coding", []), "coding"
                    )
                )
                if simple_representation:
                    output.append(simple_representation)
        # ignore non lists unless it's a FHIR object we know how to parse
        elif type(getattr(fhir_object, list_variable, None)) == FHIRReference:
            simple_representation = getattr(fhir_object, list_variable, None).display
            if simple_representation:
                output.append(simple_representation)
        elif type(getattr(fhir_object, list_variable, None)) == Meta:
            # ignore meta info
            pass
        elif type(getattr(fhir_object, list_variable, None)) == Identifier:
            simple_representation = dbgapFHIR._get_simple_text_from_fhir_object(
                getattr(fhir_object, list_variable, None), "identifier"
            )
            if simple_representation:
                output.append(simple_representation)
        # elif ... add any more FHIR objects we can simplify and parse here
        else:
            logging.debug(
                f"could not parse FHIR object `{fhir_object}` to simple text, ignoring"
            )

        return output

    def _flatten_relevant_fields(self, dictionary):
        for item in dictionary.get("Content", []):
            for item in dictionary["Content"]:
                for key, value in item.items():
                    if (
                        isinstance(value, collections.abc.Iterable)
                        and not type(value) is str
                    ):
                        if len(value) > 1:
                            logging.warning(
                                f"Content field {key} contains more than 1 value: {value} "
                                "(which is unexpected), using 1st item ONLY."
                            )
                        dictionary[key] = str(value[0])

        try:
            del dictionary["Content"]
        except KeyError:
            pass

        if dictionary.get("StudyConsents", []):
            for item in dictionary["StudyConsents"]:
                consents = item.get("StudyConsent")
                if consents:
                    dictionary["StudyConsents"] = consents

        if dictionary.get("Citers", []):
            citers = []
            for item in dictionary.get("Citers", []):
                for citer in item.get("Citer", []):
                    """
                    ex:

                    {'Title': ['A novel variational Bayes multiple locus Z-statistic '
                           'for genome-wide association studies with Bayesian '
                           'model averaging'],
                    'Url': 'https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3381972'}
                    """
                    citer_sanitized = {
                        "Title": citer.get("Title", [""])[0],
                        "Url": citer.get("Url", [""]),
                    }
                    citers.append(citer_sanitized)
            dictionary["Citers"] = citers

        for item in self.suspected_single_item_list_fields:
            if dictionary.get(item) and len(dictionary.get(item, [])) == 1:
                dictionary[item] = dictionary[item][0]
            elif dictionary.get(item) and len(dictionary.get(item, [])) > 1:
                logging.warning(
                    f"expected {item} in dict to have single "
                    f"item list. it has more: {dictionary.get(item)}"
                )

    def _remove_unecessary_fields(self, all_data):
        # remove some fields deemed unnecessary or not typically populated
        for field in self.unnecessary_fields:
            try:
                del all_data[field]
            except KeyError:
                pass

    @staticmethod
    def _capitalize_top_level_keys(all_data):
        for key, value in copy.deepcopy(all_data).items():
            capitalized_key = key[:1].upper() + key[1:]
            all_data[capitalized_key] = all_data[key]

            if key != capitalized_key:
                del all_data[key]
